# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the Asset artifacts."""

import base64

from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import IntegrityError, transaction

from debusine.assets import (
    AWSProviderAccountConfiguration,
    AWSProviderAccountCredentials,
    AWSProviderAccountData,
    AssetCategory,
    BasicAPTAuthenticationData,
    BasicAPTCredentials,
    KeyPurpose,
    SigningKeyData,
)
from debusine.db.context import context
from debusine.db.models import Asset, Scope, Workspace
from debusine.db.models.assets import (
    AssetRoles,
    AssetUsage,
    AssetUsageRole,
    AssetUsageRoleBase,
    AssetUsageRoles,
)
from debusine.db.models.scopes import ScopeRoles
from debusine.db.models.tests.utils import RolesTestCase
from debusine.db.permissioncontext import PermissionContext
from debusine.db.playground import scenarios
from debusine.tasks.models import WorkerType
from debusine.test.django import TestCase


class AssetUsageRolesTests(RolesTestCase[AssetUsageRoles]):
    """Tests for the AssetUsageRoles class."""

    roles_class = AssetUsageRoles

    def test_invariants(self) -> None:
        self.assertRolesInvariants()

    def test_from_iterable(self) -> None:
        for value, expected in (
            ([], []),
            (["signer"], [AssetUsage.Roles.SIGNER]),
            (["repository_signer"], [AssetUsage.Roles.REPOSITORY_SIGNER]),
            ([AssetUsage.Roles.SIGNER], [AssetUsage.Roles.SIGNER]),
            (
                [AssetUsage.Roles.REPOSITORY_SIGNER],
                [AssetUsage.Roles.REPOSITORY_SIGNER],
            ),
            (["signer", "repository_signer"], [AssetUsage.Roles.SIGNER]),
            (
                [
                    "signer",
                    "signer",
                    AssetUsage.Roles.SIGNER,
                    AssetUsage.Roles.REPOSITORY_SIGNER,
                ],
                [AssetUsage.Roles.SIGNER],
            ),
        ):
            with self.subTest(value=repr(value)):
                self.assertEqual(
                    AssetUsage.Roles.from_iterable(value),
                    frozenset(expected),
                )

    def test_choices(self) -> None:
        self.assertChoices()

    def test_values(self) -> None:
        role = AssetUsage.Roles.SIGNER
        self.assertEqual(role.label, "Signer")
        self.assertEqual(
            role.description,
            "Sign data using this key",
        )
        self.assertEqual(role.implied_by_scope_roles, frozenset([]))
        self.assertEqual(role.implied_by_workspace_roles, frozenset([]))
        self.assertEqual(role.implied_by_asset_usage_roles, frozenset([role]))
        self.assertTrue(role.implies(AssetUsage.Roles.REPOSITORY_SIGNER))

        role = AssetUsage.Roles.REPOSITORY_SIGNER
        self.assertEqual(role.label, "Repository signer")
        self.assertEqual(
            role.description,
            "Sign repositories using this key",
        )
        self.assertEqual(
            role.implied_by_scope_roles, frozenset([Scope.Roles.OWNER])
        )
        self.assertEqual(
            role.implied_by_workspace_roles, frozenset([Workspace.Roles.OWNER])
        )
        self.assertEqual(
            role.implied_by_asset_usage_roles,
            frozenset([role, AssetUsageRoles.SIGNER]),
        )
        self.assertFalse(role.implies(AssetUsage.Roles.SIGNER))

        role = AssetUsage.Roles.APT_AUTHENTICATOR
        self.assertEqual(role.label, "Can use APT authentication")
        self.assertEqual(
            role.description,
            "Authenticate to an external APT repository "
            "using these credentials",
        )
        self.assertEqual(
            role.implied_by_scope_roles,
            frozenset([Scope.Roles.OWNER]),
        )

        self.assertEqual(
            role.implied_by_workspace_roles,
            frozenset([Workspace.Roles.OWNER]),
        )

        self.assertEqual(
            role.implied_by_asset_usage_roles,
            frozenset([role]),
        )

    def test_init_invalid_implication(self) -> None:
        r = AssetUsageRoleBase("test", implied_by=[None])
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            r"AssetUsage roles do not support implications by None",
        ):
            r._setup()


class AssetManagerTests(TestCase):
    """Tests for AssetManager and AssetQuerySet."""

    def test_in_current_scope(self) -> None:
        """Test in_current_scope()."""
        scope2 = self.playground.get_or_create_scope("scope2")
        default_workspace = self.playground.get_default_workspace()
        self.playground.create_signing_key_asset()

        with context.local():
            context.set_permission_context_scoped(
                default_workspace.scope, PermissionContext.for_ui_anonymous()
            )
            self.assertTrue(Asset.objects.in_current_scope().exists())

        with context.local():
            context.set_permission_context_scoped(
                scope2, PermissionContext.for_ui_anonymous()
            )
            self.assertFalse(Asset.objects.in_current_scope().exists())

    def test_can_display(self) -> None:
        """Test can_display() delegates to the workspace."""
        default_workspace = self.playground.get_default_workspace()
        private_scope = self.playground.get_or_create_scope("private")
        private_workspace = self.playground.create_workspace(
            public=False, scope=private_scope
        )
        # No access to private_workspace
        user = self.playground.create_user(username="unprivileged")
        asset = self.playground.create_signing_key_asset()
        self.playground.create_signing_key_asset(workspace=private_workspace)

        with context.local():
            context.set_permission_context_scoped(
                default_workspace.scope, PermissionContext.for_ui_user(user)
            )

            self.assertEqual(Asset.objects.can_display(context.pc).count(), 1)
            self.assertEqual(
                Asset.objects.can_display(context.pc).first(), asset
            )

    def test_can_display_cloud_provider_accounts(self) -> None:
        default_workspace = self.playground.get_default_workspace()
        user = self.playground.get_default_user()
        asset = self.playground.create_asset(
            category=AssetCategory.CLOUD_PROVIDER_ACCOUNT,
            workspace=default_workspace,
            data=AWSProviderAccountData(
                name="test",
                configuration=AWSProviderAccountConfiguration(
                    region_name="test-region"
                ),
                credentials=AWSProviderAccountCredentials(
                    access_key_id="access-key", secret_access_key="secret-key"
                ),
            ).model_dump(),
            created_by=user,
        )
        self.playground.create_group_role(asset, AssetRoles.OWNER, users=[user])

        with context.local():
            context.set_permission_context_scoped(
                default_workspace.scope, PermissionContext.for_ui_user(user)
            )
            self.assertQuerySetEqual(Asset.objects.can_display(context.pc), [])

    def test_can_manage_permissions(self) -> None:
        """Test can_manage_permissions()."""
        user = self.playground.get_default_user()
        asset = self.playground.create_signing_key_asset()
        self.assertFalse(
            Asset.objects.can_manage_permissions(
                PermissionContext.for_ui_user(user)
            ).exists()
        )

        self.playground.create_group_role(asset, AssetRoles.OWNER, users=[user])
        self.assertEqual(
            Asset.objects.can_manage_permissions(
                PermissionContext.for_ui_user(user)
            ).first(),
            asset,
        )

    def test_get_by_slug_signing_key(self) -> None:
        asset = self.playground.create_signing_key_asset()
        self.assertEqual(
            Asset.objects.get_by_slug(category=asset.category, slug=asset.slug),
            asset,
        )

    def test_get_by_slug_apt_authentication_by_name(self) -> None:
        asset = self.playground.create_apt_authentication_asset(name="test-1")
        self.playground.create_apt_authentication_asset(name="test-2")
        self.assertEqual(
            Asset.objects.get_by_slug(
                category=asset.category,
                slug="test-1",
                workspace=self.playground.get_default_workspace(),
            ),
            asset,
        )

    def test_get_by_slug_apt_authentication_by_workspace_and_name(self) -> None:
        other_workspace = self.playground.create_workspace(name="other")
        asset = self.playground.create_apt_authentication_asset(
            workspace=other_workspace, name="test"
        )
        self.playground.create_apt_authentication_asset(name="test")
        self.assertEqual(
            Asset.objects.get_by_slug(
                category=asset.category,
                slug="other:test",
                workspace=self.playground.get_default_workspace(),
            ),
            asset,
        )

    def test_get_by_slug_apt_authentication_system(self) -> None:
        asset = self.playground.create_apt_authentication_asset(
            name="test", workspace=None
        )
        self.playground.create_apt_authentication_asset(name="test")
        self.assertEqual(
            Asset.objects.get_by_slug(
                category=asset.category,
                slug=":test",
            ),
            asset,
        )

    def test_get_by_slug_unknown_category(self) -> None:
        with self.assertRaisesRegex(
            ValueError, r"No slug defined for category 'unknown'"
        ):
            Asset.objects.get_by_slug(category="unknown", slug="abc123")


class AssetUsageManagerTests(TestCase):
    """Tests for AssetUsageManager."""

    scenario = scenarios.DefaultContext()

    def test_get_roles_model(self) -> None:
        self.assertEqual(AssetUsage.objects.get_roles_model(), AssetUsageRole)

    def test_with_role(self) -> None:
        asset = self.playground.create_signing_key_asset()
        asset_usage = self.playground.create_asset_usage(resource=asset)
        signers = self.playground.create_group_role(
            asset_usage, AssetUsageRoles.SIGNER, users=[self.scenario.user]
        )
        other = self.playground.create_user("other")

        for pc, expected in (
            (PermissionContext.for_ui_anonymous(), []),
            (PermissionContext.for_ui_user(other), []),
            (
                self.playground.create_pc_with_extra_groups(
                    signers, user=other
                ),
                [asset_usage],
            ),
            (self.scenario.context_ui(), [asset_usage]),
        ):
            with self.subTest(pc=str(pc)):
                self.assertQuerySetEqual(
                    AssetUsage.objects.with_role(pc, AssetUsage.Roles.SIGNER),
                    expected,
                )


class AssetTests(TestCase):
    """Tests for Asset."""

    def create_signing_key_asset(self) -> Asset:
        """Create an unsaved signing-key Asset."""
        return Asset(
            category=AssetCategory.SIGNING_KEY,
            workspace=self.playground.get_default_workspace(),
            data=SigningKeyData(
                purpose=KeyPurpose.OPENPGP_UPLOAD,
                fingerprint="ABC123",
                public_key=base64.b64encode(b"PUBLIC KEY").decode(),
                description="A Description",
            ).model_dump(),
            created_by=self.playground.get_default_user(),
        )

    def create_cloud_provider_account_asset(
        self, workspace: Workspace | None
    ) -> Asset:
        """Create an unsaved cloud-provider-account Asset."""
        return Asset(
            category=AssetCategory.CLOUD_PROVIDER_ACCOUNT,
            workspace=workspace,
            data=AWSProviderAccountData(
                name="test",
                configuration=AWSProviderAccountConfiguration(
                    region_name="test-region"
                ),
                credentials=AWSProviderAccountCredentials(
                    access_key_id="access-key", secret_access_key="secret-key"
                ),
            ).model_dump(),
            created_by=self.playground.get_default_user(),
        )

    def create_apt_authentication_asset(self, system: bool = False) -> Asset:
        """Create an unsaved apt-authentication Asset."""
        return Asset(
            category=AssetCategory.APT_AUTHENTICATION,
            workspace=(
                None if system else self.playground.get_default_workspace()
            ),
            data=BasicAPTAuthenticationData(
                name="test",
                credentials=BasicAPTCredentials(
                    username="test-username", password="test-password"
                ),
            ).model_dump(),
            created_by=self.playground.get_default_user(),
        )

    def test_signing_key_unique_constraint(self) -> None:
        """Test that signing-key assets have a unique constraint."""
        self.playground.create_signing_key_asset(fingerprint="ABC123")
        with self.assertRaises(IntegrityError):
            self.playground.create_signing_key_asset(fingerprint="ABC123")

    def test_apt_authentication_name(self) -> None:
        """apt-authentication assets must have valid names."""
        valid_names = ("secret", "foo_bar", "foo.bar", "foo+bar", "tail_")
        invalid_names = (
            "+foo",
            "-foo",
            ".foo",
            "_foo",
            "foo:bar",
            "/etc/passwd",
            "../foo",
        )

        for name in valid_names:
            with self.subTest(name=name):
                self.playground.create_apt_authentication_asset(name=name)

        for name in invalid_names:
            with (
                self.subTest(name=name),
                transaction.atomic(),
                self.assertRaises(IntegrityError),
            ):
                self.playground.create_apt_authentication_asset(name=name)

    def test_apt_authentication_unique_constraint(self) -> None:
        """Test that apt-authentication assets have a unique constraint."""
        self.playground.create_apt_authentication_asset(name="test-apt")
        with self.assertRaises(IntegrityError):
            self.playground.create_apt_authentication_asset(name="test-apt")

    def test_clean(self) -> None:
        """Test that clean() validates the model."""
        asset = self.playground.create_signing_key_asset()
        asset.clean()
        asset.data["something"] = 42
        with self.assertRaisesRegex(
            ValidationError, r"invalid asset category or data"
        ):
            asset.clean()

    def test_slug_signing_key(self) -> None:
        asset = self.create_signing_key_asset()
        self.assertEqual(asset.slug, "openpgp:ABC123")

    def test_slug_apt_authentication(self) -> None:
        asset = self.create_apt_authentication_asset()
        workspace = self.playground.get_default_workspace()
        self.assertEqual(asset.slug, f"{workspace.name}:test")

    def test_slug_apt_authentication_system(self) -> None:
        asset = self.create_apt_authentication_asset(system=True)
        self.assertEqual(asset.slug, ":test")

    def test_slug_unknown_category(self) -> None:
        asset = self.create_signing_key_asset()
        asset.category = "unknown"
        with self.assertRaisesRegex(
            NotImplementedError, r"No slug defined for category 'unknown'"
        ):
            asset.slug

    def test_can_edit(self) -> None:
        """Test that can_edit() only permits direct owners to edit."""
        asset = self.playground.create_signing_key_asset()
        self.assertFalse(asset.can_edit(PermissionContext.for_ui_anonymous()))
        user = self.playground.get_default_user()
        pc = PermissionContext.for_ui_user(user)
        self.assertFalse(asset.can_edit(pc))

        self.playground.create_group_role(asset, AssetRoles.OWNER, users=[user])
        self.assertTrue(asset.can_edit(pc))

    def test_can_create(self) -> None:
        """Test that can_create() only permits scope owners to create."""
        asset = self.create_signing_key_asset()
        assert asset.workspace

        self.assertFalse(asset.can_create(PermissionContext.for_ui_anonymous()))
        user = self.playground.get_default_user()
        pc = PermissionContext.for_ui_user(user)
        self.assertFalse(asset.can_create(pc))

        self.playground.create_group_role(
            asset.workspace.scope, ScopeRoles.OWNER, users=[user]
        )
        self.assertTrue(asset.can_create(pc))

    def test_can_create_worker(self) -> None:
        """Test that can_create() can be used by signing workers."""
        asset = self.create_signing_key_asset()
        assert asset.workspace

        self.assertFalse(asset.can_create(PermissionContext.for_ui_anonymous()))
        ext_worker_token = self.playground.create_work_request_token(
            self.playground.create_worker_task(
                mark_running=True, assign_new_worker=True
            )
        )
        sig_worker_token = self.playground.create_work_request_token(
            self.playground.create_worker_task(
                mark_running=True,
                worker=self.playground.create_worker(
                    worker_type=WorkerType.SIGNING
                ),
            )
        )
        with context.local():
            context.set_permission_context_scoped(
                asset.workspace.scope,
                PermissionContext.for_work_request(token=ext_worker_token),
            )
            self.assertFalse(asset.can_create(context.pc))

        with context.local():
            context.set_permission_context_scoped(
                asset.workspace.scope,
                PermissionContext.for_work_request(token=sig_worker_token),
            )
            self.assertTrue(asset.can_create(context.pc))

    def test_can_create_no_workspace(self) -> None:
        asset = self.create_cloud_provider_account_asset(workspace=None)
        user = self.playground.get_default_user()
        self.assertFalse(asset.can_create(PermissionContext.for_ui_user(user)))

    def test_can_create_cloud_provider_account(self) -> None:
        asset = self.create_cloud_provider_account_asset(
            workspace=self.playground.get_default_workspace()
        )
        user = self.playground.get_default_user()
        assert asset.workspace
        self.playground.create_group_role(
            asset.workspace.scope, ScopeRoles.OWNER, users=[user]
        )
        self.assertFalse(asset.can_create(PermissionContext.for_ui_user(user)))

    def test_data_model(self) -> None:
        """Test data_model validations: No data."""
        asset = self.playground.create_signing_key_asset()

        self.assertIsInstance(asset.data_model, SigningKeyData)

    def test_data_model_validation_none(self) -> None:
        """Test data_model validations: No data."""
        asset = self.create_signing_key_asset()
        asset.data = None

        with self.assertRaisesRegex(
            ValidationError, r"data must be a dictionary"
        ):
            asset.data_model

    def test_data_model_validation_category(self) -> None:
        """Test data_model validations: Unknown category."""
        asset = self.create_signing_key_asset()
        asset.category = "unknown"

        with self.assertRaisesRegex(ValidationError, r"invalid asset category"):
            asset.data_model

    def test_save_create(self) -> None:
        """Test that save requires can_create permission for creation."""
        user = self.playground.get_default_user()
        asset = self.create_signing_key_asset()
        assert asset.workspace
        self.playground.create_group_role(
            asset.workspace.scope, ScopeRoles.OWNER, users=[user]
        )
        with context.local():
            context.set_permission_context_scoped(
                asset.workspace.scope, PermissionContext.for_ui_user(user)
            )
            asset.workspace.set_current()
            self.assertTrue(asset.can_create(context.pc))

            asset.save()

    def test_save_edit(self) -> None:
        """Test that save requires can_edit permission for modification."""
        user = self.playground.get_default_user()
        asset = self.playground.create_signing_key_asset()
        assert asset.workspace
        self.playground.create_group_role(asset, AssetRoles.OWNER, users=[user])
        with context.local():
            context.set_permission_context_scoped(
                asset.workspace.scope, PermissionContext.for_ui_user(user)
            )
            asset.workspace.set_current()
            self.assertTrue(asset.can_edit(context.pc))

            asset.save()

    def test_str(self) -> None:
        asset = self.create_signing_key_asset()
        self.assertEqual(
            str(asset),
            (
                "Id: None Category: debusine:signing-key "
                "Workspace: debusine/System"
            ),
        )


class AssetUsageTests(TestCase):
    """Tests for AssetUsage."""

    scenario = scenarios.DefaultContext()

    def test_has_role(self) -> None:
        asset = self.playground.create_signing_key_asset()
        asset_usage = self.playground.create_asset_usage(resource=asset)
        signers = self.playground.create_group_role(
            asset_usage, AssetUsageRoles.SIGNER, users=[self.scenario.user]
        )
        other = self.playground.create_user("other")

        for pc, expected in (
            (PermissionContext.for_ui_anonymous(), False),
            (PermissionContext.for_ui_user(other), False),
            (
                self.playground.create_pc_with_extra_groups(
                    signers, user=other
                ),
                True,
            ),
            (self.scenario.context_ui(), True),
        ):
            with self.subTest(pc=str(pc)):
                self.assertEqual(
                    asset_usage.has_role(pc, AssetUsage.Roles.SIGNER),
                    expected,
                )

    def test_can_sign_with(self) -> None:
        user = self.scenario.user
        asset = self.playground.create_signing_key_asset()
        asset_usage = self.playground.create_asset_usage(resource=asset)
        for u, workspace_roles, asset_usage_roles, expected in (
            (AnonymousUser(), [], [], False),
            (user, [], [], False),
            (user, [Workspace.Roles.OWNER], [], False),
            (user, [], [AssetUsage.Roles.SIGNER], True),
            (user, [], [AssetUsage.Roles.REPOSITORY_SIGNER], False),
        ):
            with (
                self.subTest(
                    user=repr(u),
                    workspace_roles=repr(workspace_roles),
                    asset_usage_roles=repr(asset_usage_roles),
                ),
                self.scenario.assign_role(
                    self.scenario.workspace, *workspace_roles
                ),
                self.scenario.assign_role(asset_usage, *asset_usage_roles),
            ):
                if u.is_authenticated:
                    pc = PermissionContext.for_ui_user(u)
                else:
                    pc = PermissionContext.for_ui_anonymous()
                self.assertPermissionPredicate(
                    asset_usage.can_sign_with,
                    pc,
                    expected,
                )

    def test_can_sign_repository_with(self) -> None:
        user = self.scenario.user
        asset = self.playground.create_signing_key_asset()
        asset_usage = self.playground.create_asset_usage(resource=asset)
        for u, workspace_roles, asset_usage_roles, expected in (
            (AnonymousUser(), [], [], False),
            (user, [], [], False),
            (user, [Workspace.Roles.OWNER], [], True),
            (user, [], [AssetUsage.Roles.SIGNER], True),
            (user, [], [AssetUsage.Roles.REPOSITORY_SIGNER], True),
        ):
            with (
                self.subTest(
                    user=repr(u),
                    workspace_roles=repr(workspace_roles),
                    asset_usage_roles=repr(asset_usage_roles),
                ),
                self.scenario.assign_role(
                    self.scenario.workspace, *workspace_roles
                ),
                self.scenario.assign_role(asset_usage, *asset_usage_roles),
            ):
                if u.is_authenticated:
                    pc = PermissionContext.for_ui_user(u)
                else:
                    pc = PermissionContext.for_ui_anonymous()
                self.assertPermissionPredicate(
                    asset_usage.can_sign_repository_with, pc, expected
                )

    def test_can_use_apt_authentication_with(self) -> None:
        user = self.scenario.user
        asset = self.playground.create_apt_authentication_asset()
        asset_usage = self.playground.create_asset_usage(resource=asset)
        for u, workspace_roles, asset_usage_roles, expected in (
            (AnonymousUser(), [], [], False),
            (user, [], [], False),
            (user, [Workspace.Roles.OWNER], [], True),
            (user, [], [AssetUsage.Roles.APT_AUTHENTICATOR], True),
        ):
            with (
                self.subTest(
                    user=repr(u),
                    workspace_roles=repr(workspace_roles),
                    asset_usage_roles=repr(asset_usage_roles),
                ),
                self.scenario.assign_role(
                    self.scenario.workspace, *workspace_roles
                ),
                self.scenario.assign_role(asset_usage, *asset_usage_roles),
            ):
                if u.is_authenticated:
                    pc = PermissionContext.for_ui_user(user)
                else:
                    pc = PermissionContext.for_ui_anonymous()
                self.assertPermissionPredicate(
                    asset_usage.can_use_apt_authentication_with, pc, expected
                )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the debusine Cli WorkRequest commands."""

import contextlib
import datetime as dt
import json
import re
import stat
from collections.abc import Generator
from functools import partial
from pathlib import Path
from typing import Any, override
from unittest import mock
from unittest.mock import MagicMock

import responses
import yaml

from debusine.artifacts.models import ArtifactCategory, WorkRequestResults
from debusine.client.cli import Cli
from debusine.client.commands.base import Command
from debusine.client.commands.tests.base import BaseCliTests
from debusine.client.commands.work_requests import WorkRequestModelCommand
from debusine.client.debusine import Debusine
from debusine.client.exceptions import DebusineError
from debusine.client.models import (
    ArtifactResponse,
    WorkRequestExternalDebsignRequest,
    WorkRequestRequest,
    WorkRequestResponse,
)
from debusine.test.test_utils import (
    create_artifact_response,
    create_file_response,
    create_remote_artifact,
    create_work_request_response,
)
from debusine.utils import calculate_hash


class WorkRequestModelCommandTests(BaseCliTests):
    """Tests for :py:class:`WorkRequestModelCommand`."""

    @Command.preserve_registry()
    def _command(self) -> WorkRequestModelCommand:
        class ConcreteWorkRequestModelCommand(WorkRequestModelCommand):
            """Concrete version of WorkRequestModelCommand."""

            @override
            def run(self) -> None:
                raise NotImplementedError()

        return self.command(
            ConcreteWorkRequestModelCommand, workspace="workspace"
        )

    def test_list_rich(self) -> None:
        wr = create_work_request_response()
        command = self._command()
        webui_url = command.debusine.webui_url(wr.get_link_url("webui_self"))
        with (
            mock.patch(
                "debusine.client.commands.work_requests.Table",
                return_value=(table := mock.MagicMock()),
            ),
            mock.patch(
                "debusine.client.commands.work_requests.rich.print"
            ) as rich_print,
            mock.patch("rich.json.JSON.from_data", side_effect=lambda x: x),
        ):
            stderr, stdout = self.capture_output(
                partial(command._list_rich, [wr])
            )

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        rich_print.assert_called_once()

        self.assertEqual(table.add_row.call_count, 1)
        call = table.add_row.call_args_list[0]
        self.assertEqual(call.args[0], f"[link={webui_url}]#{wr.id}[/]")
        self.assertEqual(call.args[1], f"{wr.task_type} {wr.task_name}")
        self.assertEqual(call.args[2], wr.status)
        self.assertEqual(
            call.args[3], f"Created: {command._format_timestamp(wr.created_at)}"
        )
        self.assertEqual(call.args[4], command._format_int(wr.worker))
        self.assertEqual(call.args[5], wr.task_data)

    def test_list_rich_full_data(self) -> None:
        wr = create_work_request_response()
        wr.started_at = wr.created_at + dt.timedelta(minutes=2)
        wr.completed_at = wr.started_at + dt.timedelta(minutes=3)
        wr.result = WorkRequestResults.SUCCESS
        command = self._command()
        webui_url = command.debusine.webui_url(wr.get_link_url("webui_self"))
        with (
            mock.patch(
                "debusine.client.commands.work_requests.Table",
                return_value=(table := mock.MagicMock()),
            ),
            mock.patch(
                "debusine.client.commands.work_requests.rich.print"
            ) as rich_print,
            mock.patch("rich.json.JSON.from_data", side_effect=lambda x: x),
        ):
            stderr, stdout = self.capture_output(
                partial(command._list_rich, [wr])
            )

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        rich_print.assert_called_once()

        self.assertEqual(table.add_row.call_count, 1)
        call = table.add_row.call_args_list[0]
        self.assertEqual(call.args[0], f"[link={webui_url}]#{wr.id}[/]")
        self.assertEqual(call.args[1], f"{wr.task_type} {wr.task_name}")
        self.assertEqual(call.args[2], f"{wr.status} {wr.result}")
        self.assertEqual(
            call.args[3],
            "\n".join(
                [
                    f"Created: {command._format_timestamp(wr.created_at)}",
                    f"Started: {command._format_timestamp(wr.started_at)}",
                    f"Completed: {command._format_timestamp(wr.completed_at)}",
                ]
            ),
        )
        self.assertEqual(call.args[4], command._format_int(wr.worker))
        self.assertEqual(call.args[5], wr.task_data)

    def test_format_timestamp(self) -> None:
        command = self._command()
        for value, expected in (
            (None, None),
            (dt.datetime(2026, 1, 1, tzinfo=dt.UTC), "2026-01-01 00:00:00"),
            (
                dt.datetime(2026, 1, 1, microsecond=42, tzinfo=dt.UTC),
                "2026-01-01 00:00:00",
            ),
        ):
            with self.subTest(value=repr(value)):
                self.assertEqual(command._format_timestamp(value), expected)

    def test_format_int(self) -> None:
        command = self._command()
        for value, expected in ((None, None), (0, "0"), (42, "42")):
            with self.subTest(value=value):
                self.assertEqual(command._format_int(value), expected)

    def test_show_rich(self) -> None:
        wr = create_work_request_response()
        command = self._command()
        webui_url = command.debusine.webui_url(wr.get_link_url("webui_self"))
        with (
            mock.patch(
                "debusine.client.commands.work_requests.Table",
                return_value=(table := mock.MagicMock()),
            ),
            mock.patch(
                "debusine.client.commands.work_requests.rich.print"
            ) as rich_print,
            mock.patch("rich.json.JSON.from_data", side_effect=lambda x: x),
        ):
            stderr, stdout = self.capture_output(
                partial(command._show_rich, wr)
            )

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        rich_print.assert_called_once()

        self.assertEqual(table.add_row.call_count, 6)
        rows = [
            call_args_list.args
            for call_args_list in table.add_row.call_args_list
        ]
        self.assertEqual(
            rows[0], ("ID:", f"#{wr.id} [link={webui_url}]{webui_url}[/]")
        )
        self.assertEqual(rows[1], ("Task:", f"{wr.task_type} {wr.task_name}"))
        self.assertEqual(rows[2], ("Workspace:", f"{wr.scope}/{wr.workspace}"))
        self.assertEqual(
            rows[3], ("Created at:", command._format_timestamp(wr.created_at))
        )
        self.assertEqual(rows[4], ("Status:", wr.status))
        self.assertEqual(
            rows[5],
            ("Task data:", wr.task_data),
        )

    def test_show_rich_full_data(self) -> None:
        wr = create_work_request_response()
        wr.started_at = wr.created_at + dt.timedelta(minutes=2)
        wr.completed_at = wr.started_at + dt.timedelta(minutes=3)
        wr.duration = 5
        wr.result = WorkRequestResults.SUCCESS
        wr.worker = 42
        wr.priority_base = 2
        wr.priority_adjustment = 3
        wr.dynamic_task_data = {"dynamic": True}
        wr.artifacts = [1, 2, 3]
        command = self._command()
        webui_url = command.debusine.webui_url(wr.get_link_url("webui_self"))
        with (
            mock.patch(
                "debusine.client.commands.work_requests.Table",
                return_value=(table := mock.MagicMock()),
            ),
            mock.patch(
                "debusine.client.commands.work_requests.rich.print"
            ) as rich_print,
            mock.patch("rich.json.JSON.from_data", side_effect=lambda x: x),
        ):
            stderr, stdout = self.capture_output(
                partial(command._show_rich, wr)
            )

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        rich_print.assert_called_once()

        self.assertEqual(table.add_row.call_count, 13)
        rows = [
            call_args_list.args
            for call_args_list in table.add_row.call_args_list
        ]
        self.assertEqual(
            rows[0], ("ID:", f"#{wr.id} [link={webui_url}]{webui_url}[/]")
        )
        self.assertEqual(rows[1], ("Task:", f"{wr.task_type} {wr.task_name}"))
        self.assertEqual(rows[2], ("Workspace:", f"{wr.scope}/{wr.workspace}"))
        self.assertEqual(
            rows[3], ("Created at:", command._format_timestamp(wr.created_at))
        )
        self.assertEqual(
            rows[4], ("Started at:", command._format_timestamp(wr.started_at))
        )
        self.assertEqual(
            rows[5],
            ("Completed at:", command._format_timestamp(wr.completed_at)),
        )
        self.assertEqual(
            rows[6],
            ("Duration:", command._format_float(wr.duration)),
        )
        self.assertEqual(rows[7], ("Status:", f"{wr.status} {wr.result}"))
        self.assertEqual(
            rows[8],
            ("Worker:", command._format_int(wr.worker)),
        )
        self.assertEqual(rows[9], ("Priority:", "2 + 3"))
        self.assertEqual(
            rows[10],
            ("Task data:", wr.task_data),
        )
        self.assertEqual(
            rows[11],
            ("Dynamic task data:", wr.dynamic_task_data),
        )
        self.assertEqual(
            rows[12],
            ("Artifacts:", wr.artifacts),
        )


class CliCreateWorkRequestTests(BaseCliTests):
    """Tests for the Cli functionality related to create work requests."""

    def patch_work_request_create(self) -> MagicMock:
        """
        Patch Debusine.work_request_create.

        Does not set return_value / side_effect.
        """
        patcher_work_request_create = mock.patch.object(
            Debusine, 'work_request_create', autospec=True
        )
        mocked_work_request_create = patcher_work_request_create.start()
        self.addCleanup(patcher_work_request_create.stop)

        return mocked_work_request_create

    def test_create_work_request_invalid_task_name(self) -> None:
        """Cli parse the CLI and stdin to create a WorkRequest bad task_name."""
        mocked_work_request_create = self.patch_work_request_create()
        task_name = "task-name"
        mocked_work_request_create.side_effect = DebusineError(
            title=f"invalid {task_name}"
        )

        self.enterContext(
            self.patch_sys_stdin_read(
                "distribution: stretch\n"
                "package: http://..../package_1.2-3.dsc\n"
            )
        )

        cli = self.create_cli(["work-request", "create", task_name])

        exception = self.assertShowsError(cli.execute)
        self.assertDebusineError(exception, {"title": f"invalid {task_name}"})

    def assert_create_work_request(
        self,
        cli: Cli,
        mocked_post_work_request: MagicMock,
        work_request_request: WorkRequestRequest,
    ) -> None:
        """
        Call cli.execute() and assert the work request is created.

        Assert expected stderr/stdout and network call.
        """
        with (
            mock.patch(
                "debusine.client.commands.base.Command.feedback"
            ) as feedback,
            mock.patch(
                "debusine.client.commands.work_requests"
                ".WorkRequestModelCommand.show"
            ) as show,
        ):
            stderr, stdout = self.capture_output(cli.execute)

        work_request_id = mocked_post_work_request.return_value.id

        feedback.assert_called_once_with("Work request started:")
        show.assert_called_once()
        self.assertEqual(show.call_args.args[0].id, work_request_id)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")

        actual_work_request_request = mocked_post_work_request.mock_calls[0][1][
            1
        ]

        # Verify WorkRequestRequest is as expected
        assert actual_work_request_request == work_request_request

    def verify_create_work_request_scenario(
        self,
        cli_args: list[str],
        workspace: str = "developers",
        event_reactions: dict[str, Any] = {},
        use_stdin: bool = False,
    ) -> None:
        """
        Test create-work-request using workspace/stdin/stdin-data.

        :param cli_args: first parameter is task_name. create-work-request
          is prepended to them.
        :param workspace: workspace to use or None if not specified in the
          cli_args
        :param use_stdin: to pass the generated data: use_stdin or a file.
        """
        task_name = cli_args[0]

        mocked_post_work_request = self.patch_work_request_create()
        work_request_id = 11
        data = {
            "distribution": "stretch",
            "package": "http://..../package_1.2-3.dsc",
        }
        mocked_post_work_request.return_value = create_work_request_response(
            base_url=self.get_base_url(self.default_server),
            scope=self.servers[self.default_server]["scope"],
            workspace=workspace,
            id=work_request_id,
        )

        serialized_data = yaml.safe_dump(data)

        if use_stdin:
            self.enterContext(self.patch_sys_stdin_read(serialized_data))
            data_option = "-"
        else:
            data_file = self.create_temporary_file(
                contents=serialized_data.encode("utf-8")
            )
            data_option = str(data_file)

        if event_reactions:
            event_reactions_file = self.create_temporary_file(
                contents=yaml.safe_dump(event_reactions).encode("utf-8")
            )
            cli_args += ["--event-reactions", str(event_reactions_file)]

        cli = self.create_cli(
            ["work-request", "create", *cli_args, "--data", data_option]
        )

        expected_work_request_request = WorkRequestRequest(
            task_name=task_name,
            workspace=workspace,
            task_data=data,
            event_reactions=event_reactions,
        )
        self.assert_create_work_request(
            cli,
            mocked_post_work_request,
            expected_work_request_request,
        )

    def test_create_work_request_specific_workspace(self) -> None:
        """Cli parse the command line and use workspace."""
        workspace_name = "Testing"
        args = ["sbuild", "--workspace", workspace_name]
        self.verify_create_work_request_scenario(args, workspace=workspace_name)

    def test_create_work_request_event_reactions(self) -> None:
        """Cli parse the command line and provide even_reactions."""
        args = ["sbuild"]
        self.verify_create_work_request_scenario(
            args, event_reactions={'on_success': [{'send-notifications': {}}]}
        )

    def test_create_work_request_success_data_from_file(self) -> None:
        """Cli parse the command line and read file to create a work request."""
        args = ["sbuild"]
        self.verify_create_work_request_scenario(args)

    def test_create_work_request_success_data_from_stdin(self) -> None:
        """Cli parse the command line and stdin to create a work request."""
        args = ["sbuild"]
        self.verify_create_work_request_scenario(args, use_stdin=True)

    def test_create_work_request_data_is_empty(self) -> None:
        """Cli parse command line to create a work request with empty data."""
        empty_file = self.create_temporary_file()
        cli = self.create_cli(
            ["work-request", "create", "sbuild", "--data", str(empty_file)]
        )

        stderr, stdout = self.capture_output(
            cli.execute, assert_system_exit_code=3
        )

        self.assertEqual(
            stderr, "Error: data must be a dictionary. It is empty\n"
        )
        self.assertEqual(stdout, "")

    def test_create_work_request_yaml_errors_failed(self) -> None:
        """cli.execute() deal with different invalid task_data."""
        work_requests = [
            {
                "task_data": "test:\n  name: a-name\n"
                "    first-name: some first name",
                "comment": "yaml.safe_load raises ScannerError",
            },
            {
                "task_data": "input:\n  source_url: https://example.com\n )",
                "comment": "yaml.safe_load raises ParserError",
            },
        ]

        for work_request in work_requests:
            task_data = work_request["task_data"]
            with self.subTest(task_data), self.patch_sys_stdin_read(task_data):
                cli = self.create_cli(["work-request", "create", "task-name"])
                stderr, _ = self.capture_output(
                    cli.execute, assert_system_exit_code=3
                )

                self.assertRegex(stderr, "^Error parsing YAML:")
                self.assertRegex(
                    stderr,
                    "Fix the YAML data\n$",
                )


class CliManageWorkRequestTests(BaseCliTests):
    """Tests for CLI manage-work-request subcommand."""

    def test_no_changes(self) -> None:
        """manage-work-request fails if no changes are specified."""
        cli = self.create_cli(["work-request", "manage", "1"])

        stderr, stdout = self.capture_output(
            cli.execute, assert_system_exit_code=3
        )

        self.assertEqual(stderr, "Error: no changes specified\n")
        self.assertEqual(stdout, "")

    def test_set_priority_adjustment(self) -> None:
        """manage-work-request sets the priority adjustment."""
        debusine = self.patch_build_debusine_object()
        cli = self.create_cli(
            ["work-request", "manage", "--set-priority-adjustment", "100", "1"]
        )

        stderr, stdout = self.capture_output(cli.execute)

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        debusine.return_value.work_request_update.assert_called_once_with(
            1, priority_adjustment=100
        )


class CliRetryWorkRequestTests(BaseCliTests):
    """Tests for CLI retry-work-request subcommand."""

    def test_retry(self) -> None:
        """retry-work-request calls the retry method."""
        debusine = self.patch_build_debusine_object()
        cli = self.create_cli(["work-request", "retry", "1"])

        stderr, stdout = self.capture_output(cli.execute)

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        debusine.return_value.work_request_retry.assert_called_once_with(
            1,
        )


class CliAbortWorkRequestTests(BaseCliTests):
    """Tests for CLI abort-work-request subcommand."""

    def test_abort(self) -> None:
        """abort-work-request calls the abort method."""
        debusine = self.patch_build_debusine_object()
        cli = self.create_cli(["work-request", "abort", "1"])

        stderr, stdout = self.capture_output(cli.execute)

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        debusine.return_value.work_request_abort.assert_called_once_with(
            1,
        )


class CliWorkRequestListTests(BaseCliTests):
    """Tests for CLI list-work-requests method."""

    @override
    def setUp(self) -> None:
        self.artifact = create_artifact_response(id=1)
        super().setUp()

    def test_list_work_requests_success(self) -> None:
        """CLI lists work requests and prints them."""
        responses = [
            create_work_request_response(id=1),
            create_work_request_response(id=2),
        ]

        def work_request_iter(
            _self: Debusine,
        ) -> Generator[WorkRequestResponse]:
            yield from responses

        self.enterContext(
            mock.patch.object(
                Debusine,
                "work_request_iter",
                autospec=True,
                side_effect=work_request_iter,
            )
        )

        cli = self.create_cli(["work-request", "list"])

        with mock.patch(
            "debusine.client.commands.work_requests.List.list"
        ) as list_:
            stderr, stdout = self.capture_output(cli.execute)

        list_.assert_called_with(responses)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")


class CliWorkRequestStatusTests(BaseCliTests):
    """Tests for the Cli functionality related to work request status."""

    @override
    def setUp(self) -> None:
        self.artifact = create_artifact_response(
            id=1,
        )
        super().setUp()

    def artifact_get(
        self,
        debusine: None,  # noqa: ARG002, U100
        artifact_id: int,
    ) -> ArtifactResponse:
        """
        Return copy of self.artifact changing its id to artifact_id.

        :param debusine: unused, only to have the same function signature
          as Debusine.artifact_get() (so it can be used in a mock)
        :param artifact_id: if of the returned artifact.
        """
        artifact = self.artifact.model_copy()
        artifact.id = artifact_id
        return artifact

    def test_show_work_request_success(self) -> None:
        """Cli use Debusine to fetch a WorkRequest and prints it to stdout."""
        created_at = dt.datetime.now(tz=dt.UTC)
        work_request_id = 10
        artifact_ids = [1, 2]

        work_request_response = create_work_request_response(
            id=work_request_id,
            created_at=created_at,
            artifacts=artifact_ids,
            task_data={"architecture": "amd64"},
            status="running",
        )

        self.enterContext(
            mock.patch.object(
                Debusine,
                'work_request_get',
                autospec=True,
                return_value=work_request_response,
            )
        )
        self.enterContext(
            mock.patch.object(
                Debusine,
                "artifact_get",
                autospec=True,
                side_effect=self.artifact_get,
            )
        )

        cli = self.create_cli(["work-request", "show", str(work_request_id)])
        _, stdout = self.capture_output(cli.execute)

        expected_artifacts = [
            self.artifact_get(None, artifact_id).model_dump(mode="json")
            for artifact_id in artifact_ids
        ]

        stdout_parsed = yaml.safe_load(stdout)
        stdout_artifacts = stdout_parsed.pop("artifacts", None)
        stdout_parsed["artifacts"] = [a["id"] for a in stdout_artifacts]
        output = WorkRequestResponse(**stdout_parsed)
        self.assertEqual(output, work_request_response)
        self.assertEqual(stdout_artifacts, expected_artifacts)


class CliProvideSignatureTests(BaseCliTests):
    """Tests for the CLI `provide-signature` command."""

    @contextlib.contextmanager
    def patch_debusine_method(
        self, name: str, **kwargs: Any
    ) -> Generator[mock.MagicMock]:
        """Patch a method on :py:class:`Debusine`."""
        with mock.patch.object(
            Debusine, name, autospec=True, **kwargs
        ) as patcher:
            yield patcher

    @responses.activate
    def verify_provide_signature_scenario(
        self,
        local_changes: bool = False,
        extra_args: list[str] | None = None,
        workspace: str = "System",
    ) -> None:
        """Test a provide-signature scenario."""
        if extra_args is None:
            extra_args = []
        directory = self.create_temporary_directory()
        (tar := directory / "foo_1.0.tar.xz").write_text("tar")
        (dsc := directory / "foo_1.0.dsc").write_text("dsc")
        (buildinfo := directory / "foo_1.0_source.buildinfo").write_text(
            "buildinfo"
        )
        changes = directory / "foo_1.0_source.changes"
        self.write_changes_file(changes, [tar, dsc, buildinfo])
        bare_work_request_response = create_work_request_response(
            id=1,
            task_type="Wait",
            task_name="externaldebsign",
            workspace=workspace,
        )
        full_work_request_response = create_work_request_response(
            id=1,
            task_type="Wait",
            task_name="externaldebsign",
            dynamic_task_data={"unsigned_id": 2},
            workspace=workspace,
        )
        unsigned_artifact_response = create_artifact_response(
            id=2,
            files={
                path.name: create_file_response(
                    size=path.stat().st_size,
                    checksums={"sha256": calculate_hash(path, "sha256").hex()},
                    url=f"https://example.com/{path.name}",
                )
                for path in (tar, dsc, buildinfo, changes)
            },
            workspace=workspace,
        )
        remote_signed_artifact = create_remote_artifact(
            id=3, workspace=workspace
        )
        args = ["work-request", "sign", "1"]
        if local_changes:
            args.extend(["--local-file", str(changes)])
        if extra_args:
            args.extend(["--", *extra_args])
        cli = self.create_cli(args)

        for path in (tar, dsc, buildinfo, changes):
            responses.add(
                responses.GET,
                f"https://example.com/{path.name}",
                body=path.read_bytes(),
            )

        with (
            self.patch_debusine_method(
                "work_request_get", return_value=bare_work_request_response
            ) as mock_work_request_get,
            self.patch_debusine_method(
                "work_request_external_debsign_get",
                return_value=full_work_request_response,
            ) as mock_work_request_external_debsign_get,
            self.patch_debusine_method(
                "artifact_get", return_value=unsigned_artifact_response
            ) as mock_artifact_get,
            mock.patch("subprocess.run") as mock_run,
            self.patch_debusine_method(
                "upload_artifact", return_value=remote_signed_artifact
            ) as mock_upload_artifact,
            self.patch_debusine_method(
                "work_request_external_debsign_complete"
            ) as mock_work_request_external_debsign_complete,
        ):
            stderr, stdout = self.capture_output(cli.execute)

            if local_changes:
                self.assertEqual(stderr, "")
            else:
                self.assertRegex(
                    stderr,
                    r"\A"
                    + "\n".join(
                        [
                            (
                                rf"Artifact file downloaded: "
                                rf".*/{re.escape(path.name)}"
                            )
                            for path in (dsc, buildinfo, changes)
                        ]
                    )
                    + r"\n\Z",
                )
            self.assertEqual(stdout, "")

            # Ensure that the CLI called debusine in the right sequence.
            # The mocks are arranged a bit awkwardly since we want to allow
            # the unmocked Debusine.download_artifact_file to be called.
            debusine_instance = mock_work_request_get.call_args[0][0]
            self.assertIsInstance(debusine_instance, Debusine)
            mock_work_request_get.assert_called_once_with(debusine_instance, 1)
            mock_work_request_external_debsign_get.assert_called_once_with(
                debusine_instance, 1
            )
            mock_artifact_get.assert_called_once_with(debusine_instance, 2)
            mock_run.assert_called_once_with(
                ["debsign", "--re-sign", changes.name, *extra_args],
                cwd=mock.ANY,
                check=True,
            )
            mock_upload_artifact.assert_called_once_with(
                debusine_instance, mock.ANY, workspace=workspace
            )
            signed_local = mock_upload_artifact.call_args[0][1]
            self.assertEqual(signed_local.category, ArtifactCategory.UPLOAD)
            self.assertCountEqual(
                signed_local.files.keys(),
                [changes.name, dsc.name, buildinfo.name],
            )
            self.assertCountEqual(signed_local.remote_files.keys(), [tar.name])
            mock_work_request_external_debsign_complete.assert_called_once_with(
                debusine_instance,
                1,
                WorkRequestExternalDebsignRequest(signed_artifact=3),
            )
        if local_changes:
            for path in (tar, dsc, buildinfo, changes):
                url = f"https://example.com/{path.name}"
                responses.assert_call_count(url, 0)

    def test_debsign(self) -> None:
        """provide-signature calls debsign and posts the output."""
        self.verify_provide_signature_scenario()

    def test_debsign_local_changes(self) -> None:
        """provide-signature uses local .changes."""
        self.verify_provide_signature_scenario(local_changes=True)

    def test_debsign_with_args(self) -> None:
        """provide-signature calls debsign with args and posts the output."""
        self.verify_provide_signature_scenario(extra_args=["-kKEYID"])

    def test_wrong_work_request(self) -> None:
        """provide-signature only works for Wait/externaldebsign requests."""
        debusine = self.patch_build_debusine_object()
        debusine.return_value.work_request_get.return_value = (
            create_work_request_response(task_type="Wait", task_name="delay")
        )
        cli = self.create_cli(["work-request", "sign", "1"])

        stderr, _ = self.capture_output(cli.execute, assert_system_exit_code=3)

        self.assertEqual(
            stderr,
            "Don't know how to provide signature for Wait/delay work request\n",
        )

    def verify_provide_signature_error_scenario(
        self,
        error_regex: str,
        local_file: str | None = None,
        expect_size: dict[str, int] | None = None,
        expect_sha256: dict[str, str] | None = None,
        delete_dsc: bool = False,
    ) -> None:
        """Test a provide-signature failure scenario with local_file."""
        if expect_size is None:
            expect_size = {}
        if expect_sha256 is None:
            expect_sha256 = {}
        directory = self.create_temporary_directory()
        (tar := directory / "foo_1.0.tar.xz").write_text("tar")
        (dsc := directory / "foo_1.0.dsc").write_text("dsc")
        (buildinfo := directory / "foo_1.0_source.buildinfo").write_text(
            "buildinfo"
        )
        changes = directory / "foo_1.0_source.changes"
        self.write_changes_file(changes, [tar, dsc, buildinfo])
        bare_work_request_response = create_work_request_response(
            id=1, task_type="Wait", task_name="externaldebsign"
        )
        full_work_request_response = create_work_request_response(
            id=1,
            task_type="Wait",
            task_name="externaldebsign",
            dynamic_task_data={"unsigned_id": 2},
        )
        unsigned_artifact_response = create_artifact_response(
            id=2,
            files={
                path.name: create_file_response(
                    size=expect_size.get(path.name, path.stat().st_size),
                    checksums={
                        "sha256": expect_sha256.get(
                            path.name, calculate_hash(path, "sha256").hex()
                        )
                    },
                    url=f"https://example.com/{path.name}",
                )
                for path in (tar, dsc, buildinfo, changes)
            },
        )
        if delete_dsc:
            dsc.unlink()
        if local_file is None:
            local_file = str(changes)
        args = ["work-request", "sign", "1", "--local-file", local_file]
        cli = self.create_cli(args)

        with (
            self.patch_debusine_method(
                "work_request_get", return_value=bare_work_request_response
            ),
            self.patch_debusine_method(
                "work_request_external_debsign_get",
                return_value=full_work_request_response,
            ),
            self.patch_debusine_method(
                "artifact_get", return_value=unsigned_artifact_response
            ),
            mock.patch("subprocess.run"),
        ):
            stderr, _ = self.capture_output(
                cli.execute, assert_system_exit_code=3
            )

        self.assertRegex(stderr, error_regex)

    def test_missing_local_changes(self) -> None:
        """provide-signature raises an error for missing --local-file."""
        self.verify_provide_signature_error_scenario(
            r"^--local-file 'nonexistent\.changes' does not exist\.$",
            local_file="nonexistent.changes",
        )

    def test_wrong_local_changes(self) -> None:
        """provide-signature raises an error for the wrong --local-file."""
        directory = self.create_temporary_directory()
        (local_file := directory / "different.changes").write_text("")
        self.verify_provide_signature_error_scenario(
            (
                r"^'[^']+\/different\.changes' is not part of artifact \d+\. "
                r"Expecting 'foo_1.0_source\.changes'$"
            ),
            local_file=str(local_file),
        )

    def test_missing_local_dsc(self) -> None:
        """provide-signature raises an error for missing indirect local-file."""
        self.verify_provide_signature_error_scenario(
            r"^'[^']+\/foo_1.0\.dsc' does not exist\.$",
            delete_dsc=True,
        )

    def test_local_dsc(self) -> None:
        """provide-signature raises an error for the wrong kind of file."""
        self.verify_provide_signature_error_scenario(
            r"^--local-file 'foo\.dsc' is not a \.changes file.$",
            local_file="foo.dsc",
        )

    def test_size_mismatch(self) -> None:
        """provide-signature verifies the size of local files."""
        self.verify_provide_signature_error_scenario(
            r"^'[^']+/foo_1\.0\.dsc' size mismatch \(expected 999 bytes\)$",
            expect_size={"foo_1.0.dsc": 999},
        )

    def test_hash_mismatch(self) -> None:
        """provide-signature verifies the hashes of local files."""
        self.verify_provide_signature_error_scenario(
            (
                r"^'[^']+/foo_1\.0\.dsc' hash mismatch "
                r"\(expected sha256 = abc123\)$"
            ),
            expect_sha256={"foo_1.0.dsc": "abc123"},
        )


class CliOnWorkRequestTests(BaseCliTests):
    """Tests for Cli on-work-request-completed."""

    def test_command_does_not_exist(self) -> None:
        """Cli report that command does not exist."""
        command = "/some/file/does/not/exist"
        cli = self.create_cli(["work-request", "on-completed", command])

        stderr, stdout = self.capture_output(
            cli.execute, assert_system_exit_code=3
        )

        self.assertEqual(
            stderr, f'Error: "{command}" does not exist or is not executable\n'
        )
        self.assertEqual(stdout, "")

    def test_command_is_not_executable(self) -> None:
        """Cli report that command is not executable."""
        command = self.create_temporary_file()
        cli = self.create_cli(["work-request", "on-completed", str(command)])

        stderr, stdout = self.capture_output(
            cli.execute, assert_system_exit_code=3
        )

        self.assertEqual(
            stderr, f'Error: "{command}" does not exist or is not executable\n'
        )
        self.assertEqual(stdout, "")

    def debusine_on_work_request_completed_is_called(
        self,
        *,
        workspaces: list[str] | None = None,
        last_completed_at: Path | None = None,
        silent: bool | None = False,
        extra_args: list[str] | None = None,
    ) -> None:
        """Cli call debusine.on_work_request_completed."""
        command = self.create_temporary_file()
        command.chmod(stat.S_IXUSR | stat.S_IRUSR)

        debusine = self.patch_build_debusine_object()

        cli_args = []

        if silent:
            cli_args.extend(["--silent"])

        cli_args.extend(
            ["work-request", "on-completed", str(command), *(extra_args or [])]
        )

        if workspaces:
            cli_args.extend(["--workspace", *workspaces])

        if last_completed_at:
            cli_args.extend(["--last-completed-at", str(last_completed_at)])

        cli = self.create_cli(cli_args)

        stderr, stdout = self.capture_output(cli.execute)

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")

        debusine.return_value.on_work_request_completed.assert_called_with(
            workspaces=workspaces,
            last_completed_at=last_completed_at,
            command=[str(command), *(extra_args or [])],
            working_directory=Path.cwd(),
        )

    def test_debusine_on_work_request_completed(self) -> None:
        """Cli call debusine.on_work_request_completed."""
        self.debusine_on_work_request_completed_is_called(silent=True)

    def test_debusine_on_work_request_completed_with_extra_args(self) -> None:
        self.debusine_on_work_request_completed_is_called(
            silent=True, extra_args=["foo", "bar"]
        )

    def test_debusine_on_work_request_completed_with_workspace(self) -> None:
        """Cli call debusine.on_work_request_completed."""
        self.debusine_on_work_request_completed_is_called(workspaces=["System"])

    def test_debusine_on_work_request_completed_create_last_completed(
        self,
    ) -> None:
        """
        Cli call debusine.on_work_request_completed.

        File last_completed_at is created.
        """
        last_completed_at = self.create_temporary_file()
        last_completed_at.unlink()

        self.debusine_on_work_request_completed_is_called(
            last_completed_at=last_completed_at
        )

        expected = json.dumps({"last_completed_at": None}, indent=2) + "\n"
        self.assertEqual(last_completed_at.read_text(), expected)

    def test_debusine_on_work_request_completed_last_completed_existed(
        self,
    ) -> None:
        """Cli call debusine.on_work_request_completed."""
        contents = json.dumps({"last_completed_at": None}).encode("utf-8")
        last_completed_at = self.create_temporary_file(contents=contents)

        self.debusine_on_work_request_completed_is_called(
            last_completed_at=last_completed_at
        )

    def test_debusine_on_work_request_last_completed_no_write_permission(
        self,
    ) -> None:
        """Cli cannot write to last-completed-at file."""
        cannot_write = self.create_temporary_file()
        cannot_write.chmod(0o444)
        cli = self.create_cli(
            [
                "work-request",
                "on-completed",
                "/bin/echo",
                "--last-completed-at",
                str(cannot_write),
            ]
        )

        stderr, _ = self.capture_output(cli.execute, assert_system_exit_code=3)
        self.assertEqual(
            stderr, f'Error: write or read access denied for "{cannot_write}"\n'
        )

    def test_debusine_on_work_request_last_completed_no_read_permission(
        self,
    ) -> None:
        """Cli cannot write to last-completed-at file."""
        cannot_write = self.create_temporary_file()
        cannot_write.chmod(0o333)
        cli = self.create_cli(
            [
                "work-request",
                "on-completed",
                "/bin/echo",
                "--last-completed-at",
                str(cannot_write),
            ]
        )

        stderr, _ = self.capture_output(cli.execute, assert_system_exit_code=3)
        self.assertEqual(
            stderr, f'Error: write or read access denied for "{cannot_write}"\n'
        )

    def test_debusine_on_work_request_dir_last_completed_no_permission(
        self,
    ) -> None:
        """Cli cannot write to last-completed-at directory."""
        directory = self.create_temporary_directory()
        directory.chmod(0o000)
        cli = self.create_cli(
            [
                "work-request",
                "on-completed",
                "/bin/echo",
                "--last-completed-at",
                str(directory / "file.json"),
            ]
        )

        stderr, _ = self.capture_output(cli.execute, assert_system_exit_code=3)
        self.assertEqual(
            stderr, f'Error: write access denied for directory "{directory}"\n'
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Apply some recent ``task_data`` migrations to ``configured_task_data``."""

from django.db import migrations

from debusine.artifacts.models import TaskTypes
from debusine.db.migrations._utils import make_data_field_renamer

# Same as in 0028_rename_arch_all_build_architecture.
affected_workflows = {
    "autopkgtest",
    "blhc",
    "debdiff",
    "debian_pipeline",
    "lintian",
    "package_upload",
    "piuparts",
    "qa",
    "reverse_dependencies_autopkgtest",
    "sbuild",
}


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0030_rename_more_build_architecture"),
    ]

    operations = [
        make_data_field_renamer(
            "WorkRequest",
            "configured_task_data",
            {"task_type": TaskTypes.WORKFLOW, "task_name": "debian_pipeline"},
            [("publish_target", "suite")],
        ),
        make_data_field_renamer(
            "WorkRequest",
            "configured_task_data",
            {
                "task_type": TaskTypes.WORKFLOW,
                "task_name__in": affected_workflows,
            },
            [("arch_all_host_architecture", "arch_all_build_architecture")],
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for AWS EC2 cloud workers."""

from base64 import b64decode
from collections import UserList
from typing import Any, TYPE_CHECKING, override
from unittest import TestCase as SimpleTestCase
from unittest.mock import call, patch

from botocore import stub
from botocore.exceptions import ClientError

from debusine.assets import AWSProviderAccountData, CloudProvidersType
from debusine.db.context import context
from debusine.db.models import Asset, Worker, WorkerPool
from debusine.server.worker_pools.amazon_ec2 import AWSEC2WorkerPool, log
from debusine.server.worker_pools.cloud_init import FilesystemSetup
from debusine.server.worker_pools.exceptions import InstanceLaunchException
from debusine.server.worker_pools.models import (
    AWSEC2InstanceMarketType,
    AWSEC2InstanceRequirements,
    AWSEC2LaunchTemplate,
    AWSEC2NetworkInterface,
    MinMaxRequirements,
)
from debusine.test.django import TestCase

if TYPE_CHECKING:
    from mypy_boto3_ec2.type_defs import DescribeInstancesResultTypeDef
else:
    DescribeInstancesResultTypeDef = dict[str, Any]


class UnsortedList[LI](UserList[LI]):
    """List that compares values in any order."""

    @override
    def __eq__(self, other: object) -> bool:
        if isinstance(other, list):
            return set(self.data) == set(other)
        return super().__eq__(other)


class TestUnsortedList(SimpleTestCase):
    """Test UnsortedList."""

    def test_equal(self) -> None:
        self.assertEqual(UnsortedList([1, 2, 3]), [3, 2, 1])
        self.assertEqual(UnsortedList(['a', 'b', 'c']), ['c', 'b', 'a'])

    def test_not_equal(self) -> None:
        self.assertNotEqual(UnsortedList(['a', 'b', 'c']), ['a', 'b', 'c', 'd'])
        self.assertNotEqual(UnsortedList(['a', 'b', 'c']), 42)


class TestAWSEC2WorkerPool(TestCase):
    """Test for AWSEC2WorkerPool."""

    provider_account: Asset
    worker_pool_model: WorkerPool

    @override
    def setUp(self) -> None:
        super().setUp()
        self.provider_account = (
            self.playground.create_cloud_provider_account_asset(
                cloud_provider=CloudProvidersType.AWS
            )
        )
        self.worker_pool_model = self.playground.create_worker_pool(
            provider_account=self.provider_account
        )

    def get_worker_pool(self) -> AWSEC2WorkerPool:
        """Return a AWSEC2WorkerPool instance."""
        worker_pool = self.worker_pool_model.provider_interface
        assert isinstance(worker_pool, AWSEC2WorkerPool)
        return worker_pool

    def test_client_attributes(self) -> None:
        worker_pool = self.get_worker_pool()
        client = worker_pool.client
        self.assertEqual(client.__class__.__name__, "EC2")
        self.assertIsNotNone(client)
        provider_account_data = self.provider_account.data_model
        assert isinstance(provider_account_data, AWSProviderAccountData)
        self.assertEqual(
            client.meta.region_name,
            provider_account_data.configuration.region_name,
        )
        # Private attribute not exposed by type annotations.
        signer = getattr(client, "_request_signer")
        self.assertEqual(
            signer._credentials.access_key,
            provider_account_data.credentials.access_key_id,
        )
        self.assertEqual(
            signer._credentials.secret_key,
            provider_account_data.credentials.secret_access_key,
        )

    def test_client_endpoint_url(self) -> None:
        provider_account_data = self.provider_account.data_model
        assert isinstance(provider_account_data, AWSProviderAccountData)
        provider_account_data.configuration.ec2_endpoint_url = (
            "https://example.com/"
        )
        with context.disable_permission_checks():
            self.provider_account.data = provider_account_data.model_dump(
                exclude_unset=True
            )
            self.provider_account.save()

        worker_pool = self.get_worker_pool()
        self.assertEqual(
            worker_pool.client.meta.endpoint_url, "https://example.com/"
        )

    def launch_template(self, **kwargs: Any) -> AWSEC2LaunchTemplate:
        """Generate a test AWSEC2LaunchTemplate."""
        launch_template = AWSEC2LaunchTemplate(
            ImageId="ec2-image",
            InstanceType="t3a.medium",
        )
        return launch_template.model_copy(update=kwargs)

    def test_terminate_existing_instances(self) -> None:
        worker_pool = self.get_worker_pool()
        with stub.Stubber(worker_pool.client) as stubber:
            stubber.add_response(
                method="terminate_instances",
                expected_params={
                    "InstanceIds": ["instance-1"],
                },
                service_response={},
            )
            worker_pool._terminate_existing_instances(
                {"instance-id": "instance-1"}
            )

    def test_terminate_existing_instances_noop(self) -> None:
        worker_pool = self.get_worker_pool()
        with stub.Stubber(worker_pool.client):
            worker_pool._terminate_existing_instances(None)
            worker_pool._terminate_existing_instances({})

    def test_terminate_existing_instances_missing(self) -> None:
        worker_pool = self.get_worker_pool()
        with stub.Stubber(worker_pool.client) as stubber:
            stubber.add_client_error(
                method="terminate_instances",
                expected_params={
                    "InstanceIds": ["instance-1"],
                },
                service_error_code="InvalidInstanceID.NotFound",
                service_message="The instance ID 'instance-1' does not exist",
            )
            worker_pool._terminate_existing_instances(
                {"instance-id": "instance-1"}
            )

    def test_terminate_existing_instances_unexpected_failure(self) -> None:
        worker_pool = self.get_worker_pool()
        with (
            stub.Stubber(worker_pool.client) as stubber,
            self.assertRaisesRegex(
                ClientError, r"Some entirely unexpected failure"
            ),
        ):
            stubber.add_client_error(
                method="terminate_instances",
                expected_params={
                    "InstanceIds": ["instance-1"],
                },
                service_error_code="InternalFailure",
                service_message="Some entirely unexpected failure",
            )
            worker_pool._terminate_existing_instances(
                {"instance-id": "instance-1"}
            )

    def test_get_apt_authentication_unset(self) -> None:
        pool = self.get_worker_pool()
        self.assertIsNone(pool.get_apt_authentication())

    def test_get_apt_authentication(self) -> None:
        apt_authentication = self.playground.create_apt_authentication_asset(
            workspace=None
        )
        self.worker_pool_model.apt_authentication = apt_authentication
        self.worker_pool_model.save()

        pool = self.get_worker_pool()
        self.assertEqual(
            pool.get_apt_authentication(), apt_authentication.data_model
        )

    def test_cloud_init_config_simple(self) -> None:
        token = self.playground.create_worker_activation_token()
        config = self.get_worker_pool()._cloud_init_config(
            self.launch_template(), "worker", token
        )
        self.assertEqual(config.device_aliases, {})
        self.assertEqual(config.fs_setup, [])
        self.assertEqual(config.mounts, [])
        self.assertGreater(len(config.runcmd), 0)

    def test_cloud_init_config_swap(self) -> None:
        token = self.playground.create_worker_activation_token()
        launch_template = self.launch_template(swap_size=4)
        config = self.get_worker_pool()._cloud_init_config(
            launch_template, "worker", token
        )
        self.assertEqual(config.device_aliases, {"swap": "/dev/xvdb"})
        self.assertEqual(
            config.fs_setup, [FilesystemSetup(device="swap", filesystem="swap")]
        )
        self.assertEqual(
            config.mounts,
            [
                [
                    "swap",
                    "none",
                    "swap",
                    "sw,nofail,x-systemd.requires=cloud-init.service",
                ],
            ],
        )
        self.assertGreater(len(config.runcmd), 0)

    def test_lookup_image_root_device_name(self) -> None:
        worker_pool = self.get_worker_pool()
        with stub.Stubber(worker_pool.client) as stubber:
            stubber.add_response(
                method="describe_images",
                expected_params={"ImageIds": ["test-image"]},
                service_response={
                    "Images": [
                        {
                            "RootDeviceName": "/dev/xvda",
                            "RootDeviceType": "ebs",
                        }
                    ]
                },
            )
            root_device = worker_pool._lookup_image_root_device_name(
                "test-image"
            )
        self.assertEqual(root_device, "/dev/xvda")

    def test_lookup_image_root_device_name_no_image(self) -> None:
        worker_pool = self.get_worker_pool()
        with (
            stub.Stubber(worker_pool.client) as stubber,
            self.assertRaisesRegex(
                ValueError, r"Found 0 images instead of the expected 1"
            ),
        ):
            stubber.add_response(
                method="describe_images",
                expected_params={"ImageIds": ["test-image"]},
                service_response={"Images": []},
            )
            worker_pool._lookup_image_root_device_name("test-image")

    def test_lookup_image_root_device_name_instance_store(self) -> None:
        worker_pool = self.get_worker_pool()
        with (
            stub.Stubber(worker_pool.client) as stubber,
            self.assertRaisesRegex(
                ValueError, r"Image test-image is not EBS-backed"
            ),
        ):
            stubber.add_response(
                method="describe_images",
                expected_params={"ImageIds": ["test-image"]},
                service_response={
                    "Images": [{"RootDeviceType": "instance-store"}]
                },
            )
            worker_pool._lookup_image_root_device_name("test-image")

    def test_launch_template_data_simple(self) -> None:
        token = self.playground.create_worker_activation_token()
        launch_template = self.launch_template()
        launch_template_data = self.get_worker_pool()._launch_template_data(
            launch_template, "worker", token
        )
        # Literal copies
        self.assertEqual(
            launch_template_data["EbsOptimized"], launch_template.EbsOptimized
        )
        self.assertEqual(
            launch_template_data["ImageId"], launch_template.ImageId
        )
        self.assertEqual(
            launch_template_data["InstanceType"], launch_template.InstanceType
        )
        self.assertNotIn("InstanceRequirements", launch_template_data)
        self.assertIsNone(launch_template.KeyName)
        self.assertNotIn("KeyName", launch_template_data)
        # More complex interpretation
        self.assertEqual(launch_template_data["BlockDeviceMappings"], [])
        self.assertEqual(launch_template_data["TagSpecifications"], [])
        self.assertTrue(
            b64decode(launch_template_data["UserData"])
            .decode("utf-8")
            .startswith("#cloud-config\n")
        )

    def test_launch_template_data_instance_requirements(self) -> None:
        token = self.playground.create_worker_activation_token()
        launch_template = self.launch_template(
            InstanceType=None,
            InstanceRequirements=AWSEC2InstanceRequirements(
                VCpuCount=MinMaxRequirements(Min=1),
                MemoryMiB=MinMaxRequirements(Min=1024),
            ),
        )
        launch_template_data = self.get_worker_pool()._launch_template_data(
            launch_template, "worker", token
        )
        self.assertNotIn("InstanceType", launch_template_data)
        assert launch_template.InstanceRequirements
        self.assertEqual(
            launch_template_data["InstanceRequirements"],
            launch_template.InstanceRequirements.model_dump(exclude_unset=True),
        )

    def test_launch_template_data_ssh_key(self) -> None:
        token = self.playground.create_worker_activation_token()
        launch_template = self.launch_template(KeyName="my-key")
        launch_template_data = self.get_worker_pool()._launch_template_data(
            launch_template, "worker", token
        )
        self.assertEqual(
            launch_template_data["KeyName"], launch_template.KeyName
        )

    def test_launch_template_data_network_interfaces(self) -> None:
        token = self.playground.create_worker_activation_token()
        launch_template = self.launch_template(
            NetworkInterfaces=[
                AWSEC2NetworkInterface(
                    AssociatePublicIpAddress=True,
                    Ipv6AddressCount=1,
                )
            ],
        )
        launch_template_data = self.get_worker_pool()._launch_template_data(
            launch_template, "worker", token
        )
        self.assertEqual(
            launch_template_data["NetworkInterfaces"],
            [
                {
                    "AssociatePublicIpAddress": True,
                    "DeleteOnTermination": True,
                    "DeviceIndex": 0,
                    "Ipv6AddressCount": 1,
                },
            ],
        )

    def test_launch_template_data_resize_rootfs(self) -> None:
        token = self.playground.create_worker_activation_token()
        launch_template = self.launch_template(root_device_size=8)
        worker_pool = self.get_worker_pool()
        with patch.object(
            worker_pool,
            "_lookup_image_root_device_name",
            return_value="/dev/xvda",
        ):
            launch_template_data = worker_pool._launch_template_data(
                launch_template, "worker", token
            )
        self.assertEqual(
            launch_template_data["BlockDeviceMappings"],
            [
                {
                    "DeviceName": "/dev/xvda",
                    "Ebs": {
                        "VolumeSize": 8,
                        "DeleteOnTermination": True,
                        "VolumeType": "gp3",
                    },
                }
            ],
        )

    def test_launch_template_data_with_swap(self) -> None:
        token = self.playground.create_worker_activation_token()
        launch_template = self.launch_template(swap_size=8)
        launch_template_data = self.get_worker_pool()._launch_template_data(
            launch_template, "worker", token
        )
        self.assertEqual(
            launch_template_data["BlockDeviceMappings"],
            [
                {
                    "DeviceName": "/dev/xvdb",
                    "Ebs": {
                        "VolumeSize": 8,
                        "DeleteOnTermination": True,
                        "VolumeType": "gp3",
                    },
                }
            ],
        )

    def test_launch_template_data_with_tags(self) -> None:
        token = self.playground.create_worker_activation_token()
        launch_template = self.launch_template(tags={"foo": "bar"})
        launch_template_data = self.get_worker_pool()._launch_template_data(
            launch_template, "worker", token
        )
        self.assertEqual(
            launch_template_data["TagSpecifications"],
            [
                {
                    "ResourceType": "instance",
                    "Tags": [
                        {
                            "Key": "foo",
                            "Value": "bar",
                        },
                    ],
                },
            ],
        )

    def test_create_launch_template(self) -> None:
        worker_pool = self.get_worker_pool()
        with stub.Stubber(worker_pool.client) as stubber:
            stubber.add_response(
                method="create_launch_template",
                expected_params={
                    "LaunchTemplateName": "test-image",
                    "LaunchTemplateData": {},
                },
                service_response={
                    "LaunchTemplate": {
                        "LaunchTemplateId": "lt-id",
                        "LaunchTemplateName": "test-image",
                    },
                },
            )
            template_id = worker_pool._create_launch_template("test-image", {})
        self.assertEqual(template_id, "lt-id")

    def get_expected_create_fleet_params(
        self,
        template: str = "my-template",
        instance_market_type: str = "spot",
        worker_name: str = "cloud0",
    ) -> dict[str, Any]:
        """Get the expected parameters to create_fleet"""
        return {
            "SpotOptions": {},
            "LaunchTemplateConfigs": [
                {
                    "LaunchTemplateSpecification": {
                        "LaunchTemplateId": template,
                        "Version": "$Latest",
                    },
                }
            ],
            "TargetCapacitySpecification": {
                "TotalTargetCapacity": 1,
                "DefaultTargetCapacityType": instance_market_type,
            },
            "Type": "instant",
            "TagSpecifications": [
                {
                    "ResourceType": "instance",
                    "Tags": [{"Key": "Name", "Value": worker_name}],
                },
            ],
        }

    def test_create_fleet_spot(self) -> None:
        worker_pool = self.get_worker_pool()
        with stub.Stubber(worker_pool.client) as stubber:
            stubber.add_response(
                method="create_fleet",
                expected_params=self.get_expected_create_fleet_params(),
                service_response={
                    "Instances": [
                        {
                            "InstanceIds": ["i-id"],
                        },
                    ],
                    "Errors": [],
                },
            )
            instance_id = worker_pool._create_fleet(
                ["my-template"],
                AWSEC2InstanceMarketType.SPOT,
                worker_name="cloud0",
                max_spot_price_per_hour=None,
            )
        self.assertEqual(instance_id, "i-id")

    def test_create_fleet_spot_max_price(self) -> None:
        worker_pool = self.get_worker_pool()
        expected_params = self.get_expected_create_fleet_params()
        expected_params["SpotOptions"]["MaxTotalPrice"] = "0.1"
        with stub.Stubber(worker_pool.client) as stubber:
            stubber.add_response(
                method="create_fleet",
                expected_params=expected_params,
                service_response={
                    "Instances": [
                        {
                            "InstanceIds": ["i-id"],
                        },
                    ],
                    "Errors": [],
                },
            )
            instance_id = worker_pool._create_fleet(
                ["my-template"],
                AWSEC2InstanceMarketType.SPOT,
                worker_name="cloud0",
                max_spot_price_per_hour=0.1,
            )
        self.assertEqual(instance_id, "i-id")

    def test_create_fleet_on_demand(self) -> None:
        worker_pool = self.get_worker_pool()
        with stub.Stubber(worker_pool.client) as stubber:
            stubber.add_response(
                method="create_fleet",
                expected_params=self.get_expected_create_fleet_params(
                    instance_market_type="on-demand"
                ),
                service_response={
                    "Instances": [
                        {
                            "InstanceIds": ["i-id"],
                        },
                    ],
                    "Errors": [],
                },
            )
            instance_id = worker_pool._create_fleet(
                ["my-template"],
                AWSEC2InstanceMarketType.ON_DEMAND,
                worker_name="cloud0",
                max_spot_price_per_hour=None,
            )
        self.assertEqual(instance_id, "i-id")

    def test_create_fleet_spot_insufficient_capacity(self) -> None:
        """EC2's harmless InsufficientInstanceCapacity is logged."""
        worker_pool = self.get_worker_pool()
        with (
            stub.Stubber(worker_pool.client) as stubber,
            self.assertLogsContains(
                (
                    "WARNING:debusine.server.worker_pools.amazon_ec2:Amazon "
                    "EC2 reported InsufficientInstanceCapacity, when "
                    "attempting to launch cloud0: We have cheaper instances "
                    "in another AZ!"
                ),
                logger=log,
            ),
        ):
            stubber.add_response(
                method="create_fleet",
                expected_params=self.get_expected_create_fleet_params(),
                service_response={
                    "Instances": [
                        {
                            "InstanceIds": ["i-id"],
                        },
                    ],
                    "Errors": [
                        {
                            "ErrorCode": "InsufficientInstanceCapacity",
                            "ErrorMessage": (
                                "We have cheaper instances in another AZ!"
                            ),
                        }
                    ],
                },
            )
            instance_id = worker_pool._create_fleet(
                ["my-template"],
                AWSEC2InstanceMarketType.SPOT,
                worker_name="cloud0",
                max_spot_price_per_hour=None,
            )
        self.assertEqual(instance_id, "i-id")

    def test_create_fleet_no_templates(self) -> None:
        worker_pool = self.get_worker_pool()
        with self.assertRaisesRegex(
            ValueError, r"No launch_template_ids were provided"
        ):
            worker_pool._create_fleet(
                [],
                AWSEC2InstanceMarketType.SPOT,
                worker_name="cloud0",
                max_spot_price_per_hour=None,
            )

    def test_create_fleet_error(self) -> None:
        worker_pool = self.get_worker_pool()
        with (
            stub.Stubber(worker_pool.client) as stubber,
            self.assertRaisesRegex(
                InstanceLaunchException,
                (
                    r"Failure to launch fleet: InvalidParameterCombination: "
                    r"Foo cannot be combined with bar\."
                ),
            ),
        ):
            stubber.add_response(
                method="create_fleet",
                expected_params=self.get_expected_create_fleet_params(),
                service_response={
                    "Instances": [],
                    "Errors": [
                        {
                            "ErrorCode": "InvalidParameterCombination",
                            "ErrorMessage": "Foo cannot be combined with bar.",
                        },
                        {
                            "ErrorCode": "InvalidParameterCombination",
                            "ErrorMessage": "Foo cannot be combined with bar.",
                        },
                    ],
                },
            )
            worker_pool._create_fleet(
                ["my-template"],
                AWSEC2InstanceMarketType.SPOT,
                worker_name="cloud0",
                max_spot_price_per_hour=None,
            )

    def test_create_fleet_no_instances(self) -> None:
        worker_pool = self.get_worker_pool()
        with (
            stub.Stubber(worker_pool.client) as stubber,
            self.assertRaisesRegex(
                InstanceLaunchException, r"No instances launched"
            ),
        ):
            stubber.add_response(
                method="create_fleet",
                expected_params=self.get_expected_create_fleet_params(),
                service_response={
                    # This should never happen... But we wouldn't want to hit
                    # an IndexError
                    "Instances": [],
                    "Errors": [],
                },
            )
            worker_pool._create_fleet(
                ["my-template"],
                AWSEC2InstanceMarketType.SPOT,
                worker_name="cloud0",
                max_spot_price_per_hour=None,
            )

    def test_launch_worker(self) -> None:
        worker_pool = self.get_worker_pool()
        specs = worker_pool.specifications
        launch_template = self.launch_template()
        specs.launch_templates = [launch_template]
        worker = self.playground.create_worker(
            worker_pool=worker_pool.worker_pool
        )
        worker.worker_pool_data = {"random-data": "foo"}
        worker.instance_created_at = None
        worker.save()

        with (
            stub.Stubber(worker_pool.client) as stubber,
            patch.object(
                worker_pool, "_launch_template_data", return_value=object()
            ) as launch_template_data,
            patch.object(
                worker_pool,
                "_create_launch_template",
                return_value="template-1",
            ) as create_launch_template,
            patch.object(
                worker_pool, "_create_fleet", return_value="instance-1"
            ) as create_fleet,
            patch.object(
                worker_pool, "_terminate_existing_instances"
            ) as terminate_existing_instances,
        ):
            stubber.add_client_error(
                method="delete_launch_template",
                expected_params={
                    "LaunchTemplateName": f"debusine_worker_{worker.name}_0",
                },
                service_error_code="NotFoundException",
                service_message=(
                    f"The specified launch template, with template name "
                    f"debusine_worker_{worker.name}_0, does not exist."
                ),
            )
            stubber.add_response(
                method="delete_launch_template",
                expected_params={
                    "LaunchTemplateId": "template-1",
                },
                service_response={},
            )
            worker_pool.launch_worker(worker)

        launch_template_data.assert_called_once_with(
            launch_template, worker.name, worker.activation_token
        )
        create_launch_template.assert_called_once_with(
            f"debusine_worker_{worker.name}_0",
            launch_template_data.return_value,
        )
        create_fleet.assert_called_once_with(
            ["template-1"],
            instance_market_type=specs.instance_market_type,
            worker_name=worker.name,
            max_spot_price_per_hour=specs.max_spot_price_per_hour,
        )
        terminate_existing_instances.assert_called_once_with(
            {"random-data": "foo"}
        )
        self.assertIsNotNone(worker.instance_created_at)
        self.assertEqual(worker.worker_pool_data, {"instance-id": "instance-1"})
        assert worker.activation_token
        self.assertTrue(worker.activation_token.enabled)

    def test_terminate_worker(self) -> None:
        worker_pool = self.get_worker_pool()
        worker = self.playground.create_worker(
            worker_pool=worker_pool.worker_pool
        )
        worker.worker_pool_data = {"instance-id": "instance-1"}
        worker.save()

        with patch.object(
            worker_pool, "_terminate_existing_instances"
        ) as terminate_existing_instances:
            worker_pool.terminate_worker(worker)

        terminate_existing_instances.assert_called_once_with(
            {"instance-id": "instance-1"}
        )
        self.assertIsNone(worker.instance_created_at)
        self.assertIsNone(worker.worker_pool_data)

    def test_terminate_missing(self) -> None:
        worker_pool = self.get_worker_pool()
        workers: dict[str, Worker] = {}
        for instance_id in ("instance-1", "instance-2", "instance-3"):
            worker = self.playground.create_worker(
                worker_pool=worker_pool.worker_pool
            )
            worker.worker_pool_data = {"instance-id": instance_id}
            worker.save()
            workers[instance_id] = worker

        task = self.playground.create_worker_task(
            mark_running=True, worker=workers["instance-3"]
        )

        # Inactive pool member
        worker = self.playground.create_worker(
            worker_pool=worker_pool.worker_pool
        )
        worker.instance_created_at = None
        worker.save()

        response: DescribeInstancesResultTypeDef = {
            "Reservations": [
                {
                    "Instances": [
                        {
                            "InstanceId": "instance-1",
                            "State": {
                                "Name": "running",
                            },
                        },
                        {
                            "InstanceId": "instance-2",
                            "State": {
                                "Name": "terminated",
                            },
                        },
                    ],
                }
            ],
            "ResponseMetadata": {
                "HTTPHeaders": {},
                "HTTPStatusCode": 200,
                "RequestId": "foo",
                "RetryAttempts": 0,
            },
        }
        with (
            stub.Stubber(worker_pool.client) as stubber,
            patch.object(
                worker_pool, "_terminate_existing_instances"
            ) as terminate_existing_instances,
        ):
            stubber.add_response(
                method="describe_instances",
                expected_params={
                    "InstanceIds": UnsortedList(
                        ["instance-1", "instance-2", "instance-3"]
                    ),
                },
                service_response=response,
            )
            worker_pool.terminate_missing()

        self.assertEqual(terminate_existing_instances.call_count, 2)
        terminate_existing_instances.assert_has_calls(
            [
                call({"instance-id": "instance-2"}),
                call({"instance-id": "instance-3"}),
            ],
            any_order=True,
        )
        task.refresh_from_db()
        self.assertIsNone(task.worker)

    def test_terminate_missing_empty(self) -> None:
        worker_pool = self.get_worker_pool()
        with stub.Stubber(worker_pool.client) as stubber:
            worker_pool.terminate_missing()
        stubber.assert_no_pending_responses()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Test for the Config class."""

import contextlib
import os
import tempfile
from collections.abc import Generator
from configparser import ConfigParser
from pathlib import Path
from unittest import mock

from debusine.test import TestCase
from debusine.utils import atomic_writer
from debusine.utils.urls import DEFAULT_REQUEST_TIMEOUT
from debusine.worker.config import ConfigHandler


class ConfigHandlerTests(TestCase):
    """Tests for ConfigHandler class."""

    @classmethod
    @contextlib.contextmanager
    def _temporary_config_directory(cls) -> Generator[str]:
        """Create a temporary directory with a config.ini inside."""
        with tempfile.TemporaryDirectory(
            prefix='debusine-Worker-ConfigHandlerTests-'
        ) as temp_directory:
            cls.create_configuration(temp_directory)
            yield temp_directory

    @contextlib.contextmanager
    def assert_fails_with(self, log_message: str) -> Generator[None]:
        """Ensure the process exits with code 3 and logs an error"""
        with (
            self.assertLogsContains(
                log_message, logger="debusine.worker.config"
            ),
            self.assertRaisesSystemExit(3),
        ):
            yield

    def test_get_section_key_missing_section(self) -> None:
        """
        ConfigHandler._get_section_key with a non-existing section aborts.

        ConfigHandler writes a message to the logs and raises SystemExit.
        """
        with self.configuration_temp_directory({}) as config_dir:
            config_handler = ConfigHandler(directories=[config_dir])

            log_message = (
                'Missing required section "non-existing-section" in '
                f'the configuration file ({config_dir}/config.ini)'
            )

            with (
                self.assertLogsContains(
                    log_message, logger="debusine.worker.config"
                ),
                self.assertRaisesSystemExit(3),
            ):
                config_handler._get_section_key('non-existing-section', 'key')

    def test_get_section_key_missing_key(self) -> None:
        """
        ConfigHandler._get_section_key with a non-existing key aborts.

        ConfigHandler writes a message to the logs and raises SystemExit.
        """
        with self.configuration_temp_directory({}) as config_dir:
            config_handler = ConfigHandler(directories=[config_dir])

            log_message = (
                'Missing required key "non-existing-key" in General section in '
                f'the configuration file ({config_dir}/config.ini)'
            )

            with (
                self.assertLogsContains(
                    log_message, logger="debusine.worker.config"
                ),
                self.assertRaisesSystemExit(3),
            ):
                config_handler._get_section_key('General', 'non-existing-key')

    def test_get_section_key_section_missing_not_required(self) -> None:
        """ConfigHandler._get_section_key, non-existing section return None."""
        with self.configuration_temp_directory({}) as config_dir:
            config_handler = ConfigHandler(directories=[config_dir])

            value = config_handler._get_section_key(
                'non-existing-section', 'key', required=False
            )

            self.assertIsNone(value)

    def test_get_section_key_key_missing_not_required(self) -> None:
        """ConfigHandler._get_section_key, non-existing key return None."""
        with self.configuration_temp_directory({}) as config_dir:
            config_handler = ConfigHandler(directories=[config_dir])

            value = config_handler._get_section_key(
                'default', 'non-existing-key', required=False
            )

            self.assertIsNone(value)

    def test_simple_properties(self) -> None:
        _config = {
            "api-url": "https://example.com",
            "log-file": "/var/log/foobar",
            "log-level": "DEBUG",
        }
        with self.configuration_temp_directory(_config) as temp_directory:
            config = ConfigHandler(directories=[temp_directory])
            self.assertEqual(config.api_url, "https://example.com")

            self.assertEqual(config.log_level, "DEBUG")
        # Second access to return the cached value
        self.assertEqual(config.api_url, "https://example.com")

    def test_api_url_fails_when_https_is_required(self) -> None:
        _config = {"api-url": "http://example.com"}
        with self.configuration_temp_directory(_config) as temp_directory:
            config = ConfigHandler(
                directories=[temp_directory], require_https=True
            )
            with self.assert_fails_with("does not use HTTPS"):
                config.api_url

    def test_http_timeout_invalid(self) -> None:
        """Test worker config fallback for invalid http-timeout."""
        with self.configuration_temp_directory(
            {'http-timeout': 'abc'}
        ) as config_dir:
            config = ConfigHandler(directories=[config_dir])

            with self.assertLogs(
                logger="debusine.worker.config", level="WARNING"
            ) as cm:
                actual_timeout = config.http_timeout

        self.assertEqual(actual_timeout, DEFAULT_REQUEST_TIMEOUT)
        self.assertTrue(
            any(
                "Invalid http-timeout value in worker configuration" in msg
                for msg in cm.output
            )
        )  # pragma: no cover (can be removed post-trixie)

    def test_http_timeout_not_provided(self) -> None:
        """Test worker config when http-timeout key is missing."""
        with self.configuration_temp_directory({}) as config_dir:
            config = ConfigHandler(directories=[config_dir])
            self.assertEqual(config.http_timeout, DEFAULT_REQUEST_TIMEOUT)

    def test_choose_from_force_directory_succeeds(self) -> None:
        """ConfigHandler._choose_directory returns the passed directory."""
        with self._temporary_config_directory() as temp_directory:
            config = ConfigHandler(directories=[temp_directory])

            selected_directory = config._choose_directory(
                directories=[temp_directory]
            )

            self.assertEqual(temp_directory, selected_directory)

    def test_choose_from_force_directory_fails(self) -> None:
        """
        ConfigHandler._choose_directory aborts.

        The force_directory does not exist.
        """
        directory = '/this/directory/does/not/exist/024024024'
        log_message = (
            f"Configuration directory cannot be found in: ['{directory}']"
        )

        with self.assert_fails_with(log_message):
            ConfigHandler._choose_directory(directories=[directory])

    def test_choose_from_default_directories_fails(self) -> None:
        """
        ConfigHandler._choose_directory from the default directories aborts.

        Default directories do not exist.
        """
        directory = '/this/directory/no/exist'
        log_message = (
            f"Configuration directory cannot be found in: ['{directory}']"
        )

        with (
            mock.patch.object(
                ConfigHandler, 'default_directories', new=[directory]
            ),
            self.assert_fails_with(log_message),
        ):
            ConfigHandler()

    def test_choose_from_default_directories_succeeds(self) -> None:
        """
        ConfigHandler._choose_directory from the default directories works.

        Returns the directory that exist.
        """
        with self._temporary_config_directory() as temp_directory:
            default_directories = [
                '/this/directory/does/not/exist/242',
                temp_directory,
            ]
            with mock.patch.object(
                ConfigHandler, 'default_directories', new=default_directories
            ):
                config = ConfigHandler()

            self.assertEqual(temp_directory, config._configuration_directory)

    @classmethod
    @contextlib.contextmanager
    def _temporary_directory_with_token_file(
        cls, token: str, token_filename: str = "token"
    ) -> Generator[str]:
        with cls._temporary_config_directory() as temp_directory:
            token_path = os.path.join(temp_directory, token_filename)
            with atomic_writer(token_path, mode="w", chmod=0o600) as fd:
                fd.write(token)

            yield temp_directory

    def test_token_succeeds(self) -> None:
        """ConfigHandler.token returns the token."""
        token = 'a84b92e3c3379733bf0d2b25700c4c'

        for prop, token_filename in (
            ("token", "token"),
            ("activation_token", "activation-token"),
        ):
            with (
                self.subTest(prop=prop),
                self._temporary_directory_with_token_file(
                    token, token_filename=token_filename
                ) as temp_directory,
            ):
                config = ConfigHandler(directories=[temp_directory])

                self.assertEqual(getattr(config, prop), token)

    def test_token_strip_new_line(self) -> None:
        """
        ConfigHandler.token returns the token without a new line.

        If the token file had a newline after the token: it's stripped.
        """
        for prop, token_filename in (
            ("token", "token"),
            ("activation_token", "activation-token"),
        ):
            with (
                self.subTest(prop=prop),
                self._temporary_config_directory() as temp_directory,
            ):
                token = 'a84b92e3c3379733bf0d2b25700c4c'
                with open(
                    os.path.join(temp_directory, token_filename), 'w'
                ) as token_file:
                    token_file.write(token + '\n')

                    Path(token_file.name).chmod(0o600)

                    config = ConfigHandler(directories=[temp_directory])

                self.assertEqual(getattr(config, prop), token)

    def test_token_file_does_not_exist(self) -> None:
        """ConfigHandler.token returns None for non-existing file."""
        for prop in ("token", "activation_token"):
            with (
                self.subTest(prop=prop),
                self._temporary_config_directory() as temp_directory,
            ):
                config = ConfigHandler(directories=[temp_directory])
                # The property accesses the token file.
                self.assertIsNone(getattr(config, prop))

    def test_token_file_cannot_be_read(self) -> None:
        """Token file cannot be read (not enough permissions)."""
        for prop, token_filename in (
            ("token", "token"),
            ("activation_token", "activation-token"),
        ):
            with (
                self.subTest(prop=prop),
                self._temporary_config_directory() as temp_directory,
            ):
                token_file = os.path.join(temp_directory, token_filename)
                os.mkdir(token_file)
                Path(token_file).chmod(0o000)

                config = ConfigHandler(directories=[temp_directory])

                with (
                    self.assertLogsContains(
                        "Cannot read", logger="debusine.worker.config"
                    ),
                    self.assertRaisesSystemExit(3),
                ):
                    getattr(config, prop)

    def test_token_too_open_fails(self) -> None:
        """ConfigHandler.token aborts: too open permissions."""
        for prop, token_filename in (
            ("token", "token"),
            ("activation_token", "activation-token"),
        ):
            with (
                self.subTest(prop=prop),
                self._temporary_config_directory() as temp_directory,
            ):
                token = 'a84b92e3c3379733bf0d2b25700c4c'
                with open(
                    os.path.join(temp_directory, token_filename), 'w'
                ) as token_file:
                    token_file.write(token)

                    Path(token_file.name).chmod(0o640)

                config = ConfigHandler(directories=[temp_directory])

                log_message = (
                    f'Permission too open for {token_file.name}. '
                    'Make sure that the file is not accessible '
                    'by group or others'
                )

                with self.assert_fails_with(log_message):
                    getattr(config, prop)

    def test_token_is_cached(self) -> None:
        """ConfigHandler.token caches the result."""
        token = 'a84b92e3c3379733bf0d2b25700c4c'

        with self._temporary_directory_with_token_file(token) as temp_directory:
            config = ConfigHandler(directories=[temp_directory])
            self.assertEqual(config.token, token)

            with open(os.path.join(temp_directory, 'token'), 'w') as token_file:
                token_file.write('new-token\n')
                Path(token_file.name).chmod(0o600)

            # config.token returns the first token because it is cached
            # (and it was not written via ConfigHandler.write_token which
            # clears the cache).
            self.assertEqual(config.token, token)

    def test_write_token_clean_cache(self) -> None:
        """ConfigHandler.write_token clears ConfigHandler.token cache."""
        token = 'a84b92e3c3379733bf0d2b25700c4c'

        with self._temporary_directory_with_token_file(token) as temp_directory:
            config = ConfigHandler(directories=[temp_directory])
            self.assertEqual(config.token, token)

            config.write_token('new-token')
            # New token is returned
            self.assertEqual(config.token, 'new-token')

    def test_write_token_succeeds(self) -> None:
        """
        ConfigHandler.write_token writes the token.

        Asserts that it overwrites the existing token.
        """
        old_token = 'a84b92e3c3379733bf0d2b25700c4c'

        with self._temporary_directory_with_token_file(
            old_token
        ) as temp_directory:
            config = ConfigHandler(directories=[temp_directory])

            new_token = 'new-token'

            # In order to assert that Config.write_token truncates the file
            # the new_token needs to be shorter than the old_token
            self.assertLess(len(new_token), len(old_token))

            # write_token must overwrite the existing token
            config.write_token(new_token)

            token_file = os.path.join(temp_directory, 'token')

            with open(token_file) as file:
                # Token has been written
                self.assertEqual(file.read(), new_token)

            self.assertEqual(config.token, new_token)

            config._file_permissions_are_restricted_or_fail(token_file)

    def test_write_token_failure(self) -> None:
        """ConfigHandler.write_token cannot write to the token file."""
        with self._temporary_config_directory() as temp_directory:
            config = ConfigHandler(directories=[temp_directory])

        token_file = os.path.join(temp_directory, "token")

        with (
            self.assertLogs(logger="debusine.worker.config") as logs,
            self.assertRaisesSystemExit(3),
        ):
            # it fails because temp_directory has already been deleted
            config.write_token('some-random-token')

        self.assertRegex(
            "\n".join(logs.output),
            rf"Cannot open token file: \[Errno 2\] "
            rf"No such file or directory: '{token_file}.*\.new'",
        )

    def _mock_config_handle_property(
        self, property_to_mock: str
    ) -> mock.PropertyMock:
        """Return a PropertyMock attached to ConfigHandler class."""
        # Attach mock to the property as described in https://docs.python.org/3/library/unittest.mock.html#unittest.mock.PropertyMock  # noqa: E501
        old_property = getattr(ConfigHandler, property_to_mock)
        mocked_property = mock.PropertyMock()
        setattr(ConfigHandler, property_to_mock, mocked_property)
        self.addCleanup(setattr, ConfigHandler, property_to_mock, old_property)

        return mocked_property

    def test_validate_config_or_fail(self) -> None:
        """
        validate_config_or_fail() accesses Config properties.

        There are separate tests for the behaviour accessing the properties
        (to call Config._fail). Not re-implemented in this test to avoid
        repetition.
        """
        with self._temporary_config_directory() as temp_directory:
            config = ConfigHandler(directories=[temp_directory])

            mocked_token = self._mock_config_handle_property('token')
            mocked_activation_token = self._mock_config_handle_property(
                'activation_token'
            )
            debusine_url = self._mock_config_handle_property('api_url')

            config.validate_config_or_fail()

            mocked_token.assert_called_once_with()
            mocked_activation_token.assert_called_once_with()
            debusine_url.assert_called_once_with()

    def test_invalid_configuration_missing_debusine_url(self) -> None:
        with self.configuration_temp_directory({}) as temp_directory:
            config = ConfigHandler(directories=[temp_directory])
            log_message = (
                f'Missing required key "api-url" in General '
                f'section in the configuration file '
                f'({config.active_configuration_file})'
            )
            with self.assert_fails_with(log_message):
                config.validate_config_or_fail()

    def test_invalid_configuration_missing_general_section(self) -> None:
        """Worker() raises an exception: [General] section is not available."""
        config_temp_directory = self.create_temp_config_directory({})
        config = ConfigHandler(directories=[config_temp_directory.as_posix()])

        log_message = (
            f'Missing required section "General" in the configuration file '
            f'({config.active_configuration_file})'
        )
        with self.assert_fails_with(log_message):
            config.validate_config_or_fail()

    def test_etc_config_lower_priority(self) -> None:
        """/etc/debusine/worker is the last directory to use for config."""
        self.assertEqual(
            ConfigHandler.default_directories[-1], '/etc/debusine/worker'
        )

    def test_config_ini_cannot_be_read(self) -> None:
        """File config.ini in configuration directory cannot be read: fails."""
        with self._temporary_config_directory() as temp_directory:
            config_ini = Path(temp_directory) / 'config.ini'
            os.unlink(config_ini)
            log_message = f'Cannot read {config_ini}'

            with self.assert_fails_with(log_message):
                ConfigHandler(directories=[temp_directory])

    def test_working_directory(self) -> None:
        with self.configuration_temp_directory(
            {"working-directory": "/srv/worker"}
        ) as temp_directory:
            config = ConfigHandler(directories=[temp_directory])
            self.assertEqual(config.working_directory, Path('/srv/worker'))

    def test_working_directory_defaults_to_home_directory(self) -> None:
        with (
            self.configuration_temp_directory({}) as temp_directory,
            mock.patch.dict('os.environ', {'HOME': '/home/user'}),
        ):
            config = ConfigHandler(directories=[temp_directory])
            self.assertEqual(
                config.working_directory, Path('/home/user/workdir')
            )

    def test_working_directory_default_value_no_home(self) -> None:
        with (
            self.configuration_temp_directory({}) as temp_directory,
            mock.patch.dict('os.environ', {}, clear=True),
        ):
            config = ConfigHandler(directories=[temp_directory])
            self.assertEqual(config.working_directory, Path('/tmp'))

    def test_environment_default(self) -> None:
        with self.configuration_temp_directory({}) as temp_directory:
            config = ConfigHandler(directories=[temp_directory])
            self.assertEqual(config.environment, {})

    def test_environment_overrides(self) -> None:
        env = {"FOO": "1"}
        with self.configuration_temp_directory(
            {}, environment=env
        ) as temp_directory:
            config = ConfigHandler(directories=[temp_directory])
            self.assertEqual(config.environment, env)

    def test_environment_forces_uppercase(self) -> None:
        with self.configuration_temp_directory(
            {}, environment={"foo": "1"}
        ) as temp_directory:
            config = ConfigHandler(directories=[temp_directory])
            self.assertEqual(config.environment, {"FOO": "1"})

    @staticmethod
    def create_configuration(
        directory: str, configuration: str = '[General]\n'
    ) -> None:
        """
        Create a file named config.ini in directory with configuration_str.

        :param directory: directory to write the config.ini into. The directory
          must exist beforehand.
        :param configuration: contents of config.ini.
        """
        with open(os.path.join(directory, 'config.ini'), 'w') as config_file:
            config_file.write(configuration)

    @classmethod
    @contextlib.contextmanager
    def configuration_temp_directory(
        cls,
        general: dict[str, str],
        environment: dict[str, str] | None = None,
    ) -> Generator[str]:
        """
        Create a context to create a configuration file in a new directory.

        Yield the temporary directory name.

        :param configuration: configuration to write inside the section
          'General'.
        """
        with cls._temporary_config_directory() as temp_directory:
            config_file_name = os.path.join(temp_directory, 'config.ini')
            with open(config_file_name, 'w') as config_file:
                config_writer = ConfigParser()
                config_writer['General'] = general
                if environment:
                    config_writer['Environment'] = environment

                config_writer.write(config_file)

            yield temp_directory



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command register_pkcs11_static_key."""

import os
import random
import re
import subprocess
from typing import Any
from unittest import mock

from django.core.management import CommandError
from django.test import TestCase as DjangoTestCase
from django.test import override_settings

from debusine.django.management.tests import call_command
from debusine.signing.db.models import AuditLog, Key
from debusine.signing.models import ProtectedKeyPKCS11Static
from debusine.signing.tests import make_fake_fingerprint
from debusine.test import TestCase


@override_settings(LANGUAGE_CODE="en-us")
class RegisterPKCS11StaticKeyCommandTests(TestCase, DjangoTestCase):
    """Tests for the register_pkcs11_static_key command."""

    def test_register(self) -> None:
        """`register_pkcs11_static_key` registers a new key."""
        pkcs11_uri = "pkcs11:id=%00%01"
        public_key = random.randbytes(64)
        public_key_file = self.create_temporary_file()
        public_key_file.write_bytes(public_key)
        fingerprint = make_fake_fingerprint()

        def run(
            args: list[str | os.PathLike[str]], **kwargs: Any
        ) -> subprocess.CompletedProcess[Any]:
            fp_colons = ":".join(re.findall("..", fingerprint))
            return subprocess.CompletedProcess(
                args=args,
                returncode=0,
                stdout=f"sha256 Fingerprint={fp_colons}\n".encode(),
            )

        with mock.patch("subprocess.run", side_effect=run) as mock_run:
            call_command(
                "register_pkcs11_static_key",
                "uefi",
                pkcs11_uri,
                str(public_key_file),
                "Test key",
            )

        mock_run.assert_called_once_with(
            [
                "openssl",
                "x509",
                "-inform",
                "PEM",
                "-noout",
                "-fingerprint",
                "-sha256",
            ],
            check=True,
            input=mock.ANY,
            stdout=subprocess.PIPE,
            stderr=None,
        )
        key = Key.objects.get(purpose=Key.Purpose.UEFI, fingerprint=fingerprint)
        assert isinstance(key.stored_private_key.root, ProtectedKeyPKCS11Static)
        self.assertEqual(key.stored_private_key.root.pkcs11_uri, pkcs11_uri)
        self.assertEqual(bytes(key.public_key), public_key)
        audit_log = AuditLog.objects.latest("created_at")
        self.assertEqual(audit_log.purpose, Key.Purpose.UEFI)
        self.assertEqual(audit_log.fingerprint, fingerprint)
        self.assertEqual(audit_log.event, AuditLog.Event.REGISTER)
        self.assertEqual(audit_log.data, {"description": "Test key"})
        self.assertIsNone(audit_log.created_by_work_request_id)

    def test_register_duplicated_purpose_and_fingerprint(self) -> None:
        """`register_pkcs11_static_key` returns error: duplicate purpose/fp."""
        pkcs11_uri = "pkcs11:id=%00%01"
        public_key = random.randbytes(64)
        public_key_file = self.create_temporary_file()
        public_key_file.write_bytes(public_key)
        fingerprint = make_fake_fingerprint()
        Key.objects.create(
            purpose=Key.Purpose.UEFI,
            fingerprint=fingerprint,
            private_key={},
            public_key=public_key,
        )

        def run(
            args: list[str | os.PathLike[str]], **kwargs: Any
        ) -> subprocess.CompletedProcess[Any]:
            fp_colons = ":".join(re.findall("..", fingerprint))
            return subprocess.CompletedProcess(
                args=args,
                returncode=0,
                stdout=f"sha256 Fingerprint={fp_colons}\n".encode(),
            )

        with (
            self.assertRaisesRegex(
                CommandError,
                r"^Error creating key: "
                r"Key with this Purpose and Fingerprint already exists\.$",
            ) as exc,
            mock.patch("subprocess.run", side_effect=run),
        ):
            call_command(
                "register_pkcs11_static_key",
                "uefi",
                pkcs11_uri,
                str(public_key_file),
                "Test key",
            )

        self.assertEqual(exc.exception.returncode, 3)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""UI helpers for authentication models."""

from functools import cached_property

from django.core.exceptions import ImproperlyConfigured

from debusine.db.context import context
from debusine.db.models import Group, User
from debusine.db.models.auth import GroupMembership
from debusine.web.views.places import ResourcePlace
from debusine.web.views.ui.base import UI


class UserUI(UI[User]):
    """UI helpers for User instances."""

    @cached_property
    def place(self) -> ResourcePlace:
        """Return a place to show this user."""
        if full_name := self.instance.get_full_name():
            description = f"User {full_name} <{self.instance.username}>"
        else:
            description = f"User <{self.instance.username}>"
        place = ResourcePlace(
            title=self.instance.get_full_name() or self.instance.username,
            breadcrumb=self.instance.username,
            url=self.instance.get_absolute_url(),
            description=description,
        )
        if context.scope is not None:
            place.parent = context.scope.ui(self.request).place
        return place


class GroupUI(UI[Group]):
    """UI helpers for Group instances."""

    @cached_property
    def place(self) -> ResourcePlace:
        """Return a place to show this group."""
        if not self.request.user.is_authenticated:
            raise ImproperlyConfigured(
                "GroupUI.place invoked by an anonymous user"
            )

        scope_ui = context.require_scope().ui(self.request)
        return ResourcePlace(
            title=f"Group {self.instance}",
            breadcrumb=self.instance.name_without_scope,
            url=self.instance.get_absolute_url(),
            parent=scope_ui.place_group_list,
        )

    @cached_property
    def place_audit_log(self) -> ResourcePlace:
        """Return a place to show this group's audit log."""
        return ResourcePlace(
            title=f"Audit log for {self.instance}",
            breadcrumb="audit log",
            url=self.instance.get_absolute_url_audit_log(),
            parent=self.place,
        )


class GroupMembershipUI(UI[GroupMembership]):
    """UI helpers for GroupMembership instances."""

    @cached_property
    def place_update(self) -> ResourcePlace:
        """Return a place to update this group membership."""
        group_ui = self.instance.group.ui(self.request)
        return ResourcePlace(
            title=f"Manage membership of {self.instance.user}"
            f" in {self.instance.group}",
            breadcrumb=str(self.instance),
            url=self.instance.get_absolute_url_update(),
            parent=group_ui.place,
        )

    @cached_property
    def place_remove(self) -> ResourcePlace:
        """Return a place to remove this group membership."""
        group_ui = self.instance.group.ui(self.request)
        return ResourcePlace(
            title=f"Remove {self.instance.user} from {self.instance.group}",
            breadcrumb=str(self.instance),
            url=self.instance.get_absolute_url_remove(),
            parent=group_ui.place,
        )
